import re
from collections.abc import Iterator
from logging import error

from stub_generator.stub_generator import normalize_name


def get_argument_names(arguments, is_class):
    counts = {}
    names = []

    types = [x[0] for x in arguments]

    if is_class:
        arguments = arguments[1:]

    arg_names = [normalize_name(tp) for tp in arguments]
    for tp, arg_name in zip(arguments, arg_names):
        if arg_names.count(arg_name) == 1:
            suffix = ""
        else:
            if arg_name in counts:
                counts[arg_name] += 1
            else:
                counts[arg_name] = 1
            suffix = str(counts[arg_name])
        names.append(f"{arg_name}{suffix}")
    if is_class:
        names.insert(0, "self")
    return names, types


def _normalize_rtype(rtype):
    if rtype == "iterator":
        return "Iterator"
    return rtype


def _parse_signature(line):
    expre = re.compile(r"(\w+)\((.*)\) -> (\w+)")
    name, args, rtype = expre.match(line).group(1, 2, 3)
    args = tuple(re.findall(r"\((\w+)\) *(\w+)", args))
    return name, args, rtype


class Docs:
    def __init__(self, text, indent, is_class=False):  # noqa: C901
        self.indent = indent
        self.is_class = is_class

        if not text:
            self._set_unknown()
            return

        self.text = text
        lines = [x.strip() for x in self.text.split("\n")]

        name, self.resources = self._get_resources(lines)
        args, rtypes, infos = zip(*self.resources)
        self.rtype = _normalize_rtype(self._get_final_rtype(name, rtypes))

        self._header = self._make_doclines(infos, name)

        self.args_sets = self._get_argument_sets(args, is_class)

    def _get_argument_sets(self, args, is_class):
        args_sets = []
        for argument_set in args:
            names, types = get_argument_names(argument_set, is_class)
            args_sets.append(list(zip(names, types)))
        return args_sets

    def _make_doclines(self, infos, name):
        """
        Make the Python docstring from multiple infos.

        Cut of first and last string if they are empty.
        We cant cut off all empty lines, because it can be inside docstring.
        """
        doc_lines = []
        for doc_part in infos:
            if not doc_part:
                continue
            else:
                # cut first and last empty strings
                if not doc_part[0]:
                    doc_part = doc_part[1:]
                if not doc_part:
                    continue
                if not doc_part[-1]:
                    doc_part = doc_part[:-1]
                doc_lines.append("\n".join(doc_part))

        if len(doc_lines) > 1:
            error("Too many doclines for %s %s", name, doc_lines)

        return doc_lines

    def _get_final_rtype(self, name, rtypes):
        if len(set(rtypes)) != 1:
            error("[%s] Different rtypes", name)
        return rtypes[0]

    def _get_resources(self, lines):
        res = []
        name, args, rtype = _parse_signature(lines[0])
        res.append((args, rtype, []))
        for line in lines[1:]:
            if self._is_extra_signature(line, name):
                name, args, rtype = _parse_signature(line)
                res.append((args, rtype, []))
            else:
                res[-1][2].append(line)
        return name, res

    def _set_unknown(self):
        self.rtype = "unknown"
        self.args = ["*args"]
        self._header = ""

    def _is_extra_signature(self, line, name):
        return line.startswith(f"{name}(")

    def _format_arg_string(self, args):
        arg_string = ", ".join(args)

        # Black limit is 120, this includes
        # function definition including name and return type
        # Lets guess the best option
        # Adding trailing coma will stop black from reformatting lines,
        # so in case of long names we could just reduce this number.
        arg_size = 90

        if len(arg_string) <= arg_size:
            return arg_string
        else:
            if self.is_class:
                indent = " " * 8
            else:
                indent = " " * 4
            result = ["\n"]
            for arg in args:
                result.append(f"{indent}{arg},\n")
            result.append(indent[:-4])
            return "".join(result)

    def get_argument_strings(self) -> Iterator[str]:
        for arg_set in self.args_sets:
            if self.is_class:
                args = ["self"]
            else:
                args = []
            args.extend(f"{arg_name}: {arg_type}" for arg_name, arg_type in arg_set[self.is_class :])
            yield self._format_arg_string(args)

    def get_doc_string(self):
        doc = []
        if self._header:
            doc.append('"""')
            doc.extend(self._header)
            doc.append('"""')

        return "\n".join("{}{}".format(" " * 4 * self.indent if x else "", x) for x in doc)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_INDUSTRY_CENTER_III",
    description="PRO_INDUSTRY_CENTER_III_DESC",
    short_description="BUILDING_REFINE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=700 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_INDUSTRY_CENTER_II"],
    graphic="icons/tech/industrial_centre_iii.png",
)



import altair as alt
import pandas as pd
import streamlit as st

from tabs.widgets import colored_span, to_hex_color


def plot_for_attribute(ais_data, attribute):
    st.header(f"Distribution of the {attribute}", anchor=attribute)

    source, empires, theme_colors = get_source(ais_data, attribute)

    selection = alt.selection_multi(fields=["empire_id"], bind="legend")

    chart = (
        alt.Chart(source)
        .mark_line(interpolate="basis")
        .encode(
            x=alt.X("turn", axis=alt.Axis(grid=True, tickMinStep=1)),
            y=f"{attribute}:Q",
            color=alt.Color("empire_id:N", sort=empires),
            opacity=alt.condition(selection, alt.value(1), alt.value(0.2)),
        )
        .configure_range(category=alt.RangeScheme(theme_colors))
        .add_selection(selection)
        .interactive()
    )

    st.altair_chart(chart, use_container_width=True)


def get_source(ais_data, attribute):
    empires = [str(item["empire_id"]) for item in ais_data]
    colors = sorted((item["empire_id"], item["color"]) for item in ais_data)
    theme_colors = [to_hex_color(color) for _, color in colors]
    plot_data = {}
    for item in ais_data:
        empire_id = str(item["empire_id"])

        empire_dict = plot_data.setdefault(empire_id, {})
        for turn_info in item["turns"]:
            turn = turn_info["turn"]
            value = turn_info[attribute]
            empire_dict[turn] = value
    wide_format = pd.DataFrame(plot_data).rename_axis("turn").reset_index()
    long_format = wide_format.melt("turn", var_name="empire_id", value_name=attribute)
    return long_format, empires, theme_colors


def show_legend(ais_data):
    with st.expander("Legend", expanded=False):
        body = []
        for item in ais_data:
            empire_id = item["empire_id"]
            player = item["player_name"]
            span = colored_span(player, to_hex_color(item["color"]))
            body.append(f"""- **{empire_id}**: {span}""")
        st.markdown("\n".join(body), unsafe_allow_html=True)


def draw_plots(ais_data):
    show_legend(ais_data)
    plot_params = ["PP", "RP", "SHIP_CONT"]
    for param in plot_params:
        plot_for_attribute(ais_data, attribute=param)



from logging import debug, warning

from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()

import freeorion as fo
from operator import itemgetter

from empires import home_system_layout
from util import unique_product


def home_system_team_core(home_systems: list[int], teams: list[tuple[int, int]]) -> dict[int, int]:
    """
    Choose core for teams which is a list of pairs team id and count of empires in the team.
    Returns map from team to core home system.
    """
    if not teams:
        return {}
    debug("Teams: %s", teams)
    # sort all home systems by distance
    home_systems_distances = {}
    for hs1, hs2 in unique_product(home_systems, home_systems):
        dist = fo.jump_distance(hs1, hs2)
        home_systems_distances[(hs1, hs2)] = dist
    home_systems_sorted = sorted(home_systems_distances.items(), key=itemgetter(1), reverse=True)
    debug("Home systems sorted: %s", home_systems_sorted)

    result = {}
    if not home_systems_sorted:
        pass
    elif len(teams) == 1:
        first_team = teams[0][0]
        first_of_most_distant_systems = home_systems_sorted[0][0][0]
        result[first_team] = first_of_most_distant_systems
    else:
        first_team = teams[0][0]
        first_of_most_distant_systems = home_systems_sorted[0][0][0]
        second_team = teams[1][0]
        second_of_most_distant_systems = home_systems_sorted[0][0][1]
        result[first_team] = first_of_most_distant_systems
        result[second_team] = second_of_most_distant_systems
        if len(teams) > 2:
            warning("Teamed placement poorly implemented for %d teams", len(teams))
    return result


def place_teams_layout(  # noqa: C901
    layout: dict[int, list[int]], cores: dict[int, int], placement_teams: list[int]
) -> dict[int, int]:
    """
    Place teams on home systems layout.
    Returns map from home system to team.
    """
    # set team cores
    left_home_systems = set(layout.keys())
    result = {}
    for team, hs in cores.items():
        result[hs] = team
        left_home_systems.remove(hs)
    # for each team search for home system
    for team in placement_teams:
        # search such home system
        # 1. with maximum same team neighbors
        choose_hs = set()
        neighbors_count = None
        for hs in left_home_systems:
            cnt = len([n for n in layout[hs] if result.get(n, -1) == team])
            if neighbors_count is None or cnt > neighbors_count:
                choose_hs = {hs}
                neighbors_count = cnt
            elif cnt == neighbors_count:
                choose_hs.add(hs)

        # 2. with maximum jump distance from other teams
        if len(choose_hs) > 1:
            choose_hs2 = set()
            max_dist_to_enemy = None
            for hs in choose_hs:
                dist_to_enemy = None
                for hs2, team2 in result.items():
                    if team2 != team:
                        dist = fo.jump_distance(hs, hs2)
                        if dist_to_enemy is None or dist < dist_to_enemy:
                            dist_to_enemy = dist
                if max_dist_to_enemy is None or dist_to_enemy > max_dist_to_enemy:
                    max_dist_to_enemy = dist_to_enemy
                    choose_hs2 = {hs}
                elif dist_to_enemy == max_dist_to_enemy:
                    choose_hs2.add(hs)
            choose_hs = choose_hs2

        if choose_hs:
            hs = choose_hs.pop()
            result[hs] = team
            left_home_systems.remove(hs)
    return result.items()


def place_teams(home_systems: list[int], systems: list[int], teams: dict[int, int]) -> dict[int, int]:
    """
    Place teams on home systems layout.
    Returns map from home system to team.
    """
    # choose team's core home system
    home_system_teams = home_system_team_core(home_systems, sorted(teams.items(), key=itemgetter(1), reverse=True))
    debug("Home systems team core: %s", home_system_teams)
    # choose order of placing teams exclude already placed cores
    placement_list = []
    for team in home_system_teams.keys():
        placement_list += [(i, team) for i in range(teams[team] - (1 if team in home_system_teams.keys() else 0))]
    placement_list.sort(reverse=True)
    placement_teams = [i[1] for i in placement_list]
    debug("Placement teams: %s", placement_teams)
    # calculate placement based on teams
    layout = home_system_layout(home_systems, systems)
    debug("Home systems layout: %s", layout)
    return place_teams_layout(layout, home_system_teams, placement_teams)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_STARGATE",
    description="CON_STARGATE_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=400 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=[
        "LRN_MIND_VOID",
        "LRN_SPATIAL_DISTORT_GEN",
    ],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_STARGATE"),
    ],
    graphic="icons/building/stargate.png",
)



from dataclasses import dataclass, field

from CombatRatingsAI import ShipCombatStats
from common.fo_typing import AttackCount, AttackDamage


@dataclass
class DesignStats:
    attacks: dict[AttackDamage, AttackCount] = field(default_factory=dict)
    structure: float = 0
    shields: float = 0
    fuel: float = 0
    speed: float = 0
    stealth: float = 0
    detection: float = 0
    troops: float = 0
    colonisation: int = -1  # -1 since 0 indicates an outpost (capacity = 0)
    fuel_per_turn: float = 0
    organic_growth: float = 0
    maximum_organic_growth: float = 0
    repair_per_turn: float = 0
    asteroid_stealth: float = 0
    solar_stealth: float = 0
    fighter_capacity: float = 0
    fighter_launch_rate: float = 0
    fighter_damage: float = 0
    flak_shots: float = 0
    has_interceptors: bool = False
    damage_vs_planets: float = 0
    has_bomber: bool = False
    shield_type: str = None

    def convert_to_combat_stats(self):
        """Return a tuple as expected by CombatRatingsAI"""
        return ShipCombatStats(
            attacks=self.attacks,
            structure=self.structure,
            shields=self.shields,
            fighter_capacity=self.fighter_capacity,
            fighter_launch_rate=self.fighter_launch_rate,
            fighter_damage=self.fighter_damage,
            flak_shots=self.flak_shots,
            has_interceptors=self.has_interceptors,
            damage_vs_planets=self.damage_vs_planets,
            has_bomber=self.has_bomber,
        )



import pytest

import TechsListsAI


def get_tech_lists():
    techs = []
    for attr in dir(TechsListsAI):
        val = getattr(TechsListsAI, attr)
        if isinstance(val, type) and issubclass(val, TechsListsAI.TechGroup):
            techs.append(val)
    return techs


@pytest.fixture(scope="module", params=get_tech_lists())
def tech_group_class(request):
    yield request.param


def test_tech_group_integrity(tech_group_class):
    errors = tech_group_class().get_errors()
    assert not errors


@pytest.fixture()
def tech_list():
    class TestGroup(TechsListsAI.TechGroup):
        def __init__(self):
            super().__init__()
            self.economy.extend(
                [
                    "ECO_1",
                    "ECO_2",
                ]
            )
            self.weapon.extend(
                [
                    "WEP_1",
                    "WEP_2",
                ]
            )
            self.defense.extend(
                [
                    "DEF_1",
                    "DEF_2",
                ]
            )
            self.hull.extend(
                [
                    "HULL_1",
                    "HULL_2",
                ]
            )
            self.enqueue()

    return TestGroup()


def test_adding_rest_adds_them_in_order(tech_list):
    tech_list.enqueue()
    assert not tech_list.get_errors()

    expected = [
        "ECO_1",
        "WEP_1",
        "DEF_1",
        "HULL_1",
        "ECO_2",
        "WEP_2",
        "DEF_2",
        "HULL_2",
    ]
    assert expected == tech_list.get_techs()



from character.character_module import Aggression, Character, Trait
from character.character_strings_module import (
    _make_aggression_based_function,
    get_trait_name_aggression,
)


def test_make_aggression_based_function():
    """Check that the table adds the correct prefixes, suffixes, runs the postfix function,
    has the UNKNOWN value entry and is the correct length.
    """
    table = _make_aggression_based_function("prefix", lambda x: "post_proc_%s" % x)

    character = Character([Aggression(0)])
    assert table(character) == "post_proc_prefix_BEGINNER"

    character = Character([Aggression(5)])
    assert table(character) == "post_proc_prefix_MANIACAL"

    character = Character([Trait()])
    assert table(character) == "post_proc_UNKNOWN_VALUE_SYMBOL"


def test_make_aggression_based_function_no_post_func():
    """Check that the table doesn't run the post function if it doesn't exist"""
    table = _make_aggression_based_function("prefix")

    character = Character([Aggression(0)])
    assert table(character) == "prefix_BEGINNER"


class UnknownTrait(Trait):
    pass


def test_get_trait_name_aggression():
    """Check that asking for an unknown trait returns None"""

    character = Character([Aggression(1)])
    assert get_trait_name_aggression(character) == "UserString GSETUP_TURTLE"

    nonaggressive_character = Character([UnknownTrait()])
    assert get_trait_name_aggression(nonaggressive_character) == "UserString UNKNOWN_VALUE_SYMBOL"



from enum import IntEnum


class PriorityType(IntEnum):
    RESOURCE_GROWTH = 1  # obsolete, but removing it here breaks the save file
    RESOURCE_PRODUCTION = 2
    RESOURCE_RESEARCH = 3
    RESOURCE_TRADE = 4  # obsolete, but removing it here breaks the save file
    RESOURCE_CONSTRUCTION = 5  # obsolete, but removing it here breaks the save file
    PRODUCTION_EXPLORATION = 6
    PRODUCTION_OUTPOST = 7
    PRODUCTION_COLONISATION = 8
    PRODUCTION_INVASION = 9
    PRODUCTION_MILITARY = 10
    PRODUCTION_BUILDINGS = 11
    PRODUCTION_ORBITAL_DEFENSE = 19
    PRODUCTION_ORBITAL_INVASION = 20
    PRODUCTION_ORBITAL_OUTPOST = 21
    PRODUCTION_ORBITAL_COLONISATION = 22
    RESOURCE_INFLUENCE = 23


def get_priority_resource_types():
    return [
        PriorityType.RESOURCE_PRODUCTION,
        PriorityType.RESOURCE_RESEARCH,
        PriorityType.RESOURCE_INFLUENCE,
    ]


def get_priority_production_types():
    return [
        PriorityType.PRODUCTION_EXPLORATION,
        PriorityType.PRODUCTION_OUTPOST,
        PriorityType.PRODUCTION_COLONISATION,
        PriorityType.PRODUCTION_INVASION,
        PriorityType.PRODUCTION_MILITARY,
        PriorityType.PRODUCTION_BUILDINGS,
    ]


class MissionType(IntEnum):
    OUTPOST = 1
    COLONISATION = 2
    EXPLORATION = 5
    INVASION = 9
    MILITARY = 10
    # mostly same as MILITARY, but waits for system removal from all targeted system lists
    # (invasion, colonization, outpost, blockade) before clearing
    SECURE = 11
    ORBITAL_DEFENSE = 12
    ORBITAL_INVASION = 13
    ORBITAL_OUTPOST = 14
    # ORBITAL_COLONISATION = 15 Not implemented yet
    PROTECT_REGION = 16


class ShipRoleType(IntEnum):  # this is also used in determining fleetRoles
    INVALID = -1
    MILITARY_ATTACK = 1
    CIVILIAN_EXPLORATION = 4
    CIVILIAN_COLONISATION = 5
    CIVILIAN_OUTPOST = 6
    MILITARY_INVASION = 7
    MILITARY = 8
    BASE_DEFENSE = 9
    BASE_INVASION = 10
    BASE_OUTPOST = 11
    BASE_COLONISATION = 12


class EmpireProductionTypes(IntEnum):
    BT_BUILDING = 1  # a Building object is being built
    BT_SHIP = 2  # a Ship object is being built


class FocusType:
    FOCUS_PROTECTION = "FOCUS_PROTECTION"
    FOCUS_GROWTH = "FOCUS_GROWTH"
    FOCUS_INDUSTRY = "FOCUS_INDUSTRY"
    FOCUS_RESEARCH = "FOCUS_RESEARCH"
    FOCUS_INFLUENCE = "FOCUS_INFLUENCE"


class EmpireMeters:
    DETECTION_STRENGTH = "METER_DETECTION_STRENGTH"



import freeOrionAIInterface as fo

import AIDependencies
from buildings import BuildingType
from colonization.colony_score import debug_rating
from freeorion_tools import get_named_real, get_species_industry, tech_soon_available
from freeorion_tools.bonus_calculation import Bonus
from freeorion_tools.caching import cache_for_current_turn
from turn_state import have_honeycomb


def calculate_production(planet: fo.planet, species: fo.species, max_population: float, stability: float) -> float:
    """
    Calculate how much PP the planet's population could generate with industry focus.
    This only considers values that actually rely on industry focus, those that do not are handled by
    calculate_planet_colonization_rating._rate_focus_independent.
    """
    if stability <= 0.0:
        return 0.0

    bonus_modified = _get_production_bonus_modified(planet, stability)
    skill_multiplier = get_species_industry(species.name)
    bonus_by_policy = _get_production_bonus_mod_by_policy(stability)
    policy_multiplier = _get_policy_multiplier(stability)
    bonus_unmodified = _get_production_bonus_unmodified(planet, stability)
    bonus_flat = _get_production_flat(planet, stability)
    per_population = (
        (AIDependencies.INDUSTRY_PER_POP + bonus_modified) * skill_multiplier + bonus_by_policy
    ) * policy_multiplier + bonus_unmodified
    result = max_population * per_population + bonus_flat
    debug_rating(
        f"calculate_production pop={max_population:.2f}, st={stability:.2f}, b1={bonus_modified:.2f}, "
        f"m1={skill_multiplier:.2f}, b2={bonus_by_policy:.2f}, m2={policy_multiplier:.2f}, "
        f"b3={bonus_unmodified:.2f}, flat={bonus_flat:.2f} -> {result:.2f}"
    )
    return result


@cache_for_current_turn
def _get_modified_industry_bonuses() -> list[Bonus]:
    """
    Get list of per-population bonuses which are added before multiplication with the species skill value.
    """
    # TBD check connection!?
    have_center = bool(BuildingType.INDUSTRY_CENTER.built_or_queued_at())
    centre_bonus1 = get_named_real("BLD_INDUSTRY_CENTER_1_TARGET_INDUSTRY_PERPOP")
    centre_bonus2 = get_named_real("BLD_INDUSTRY_CENTER_2_TARGET_INDUSTRY_PERPOP")
    centre_bonus3 = get_named_real("BLD_INDUSTRY_CENTER_3_TARGET_INDUSTRY_PERPOP")
    return [
        Bonus(
            tech_soon_available("PRO_FUSION_GEN", 3),
            get_named_real("PRO_FUSION_GEN_MIN_STABILITY"),
            get_named_real("PRO_FUSION_GEN_TARGET_INDUSTRY_PERPOP"),
        ),
        Bonus(
            tech_soon_available("GRO_ENERGY_META", 3),
            get_named_real("GRO_ENERGY_META_MIN_STABILITY"),
            get_named_real("GRO_ENERGY_META_TARGET_INDUSTRY_PERPOP"),
        ),
        # special case: these are not cumulative
        Bonus(
            have_center and tech_soon_available("PRO_INDUSTRY_CENTER_III", 1),  # expensive research
            get_named_real("BLD_INDUSTRY_CENTER_3_MIN_STABILITY"),
            centre_bonus3 - centre_bonus2,
        ),
        Bonus(
            have_center and tech_soon_available("PRO_INDUSTRY_CENTER_II", 2),
            get_named_real("BLD_INDUSTRY_CENTER_2_MIN_STABILITY"),
            centre_bonus2 - centre_bonus1,
        ),
        Bonus(
            # normally we have the tech when we have the building, but we could have conquered one
            have_center and tech_soon_available("PRO_INDUSTRY_CENTER_II", 3),
            get_named_real("BLD_INDUSTRY_CENTER_1_MIN_STABILITY"),
            centre_bonus1,
        ),
    ]


def _get_production_bonus_modified(planet: fo.planet, stability: float) -> float:
    """
    Calculate bonus production per population which would be added before multiplication with the species skill value.
    """
    specials_bonus = sum(
        AIDependencies.INDUSTRY_PER_POP for s in planet.specials if s in AIDependencies.industry_boost_specials_modified
    )
    return specials_bonus + sum(bonus.get_bonus(stability) for bonus in _get_modified_industry_bonuses())


def _get_policy_multiplier(stability) -> float:
    if fo.getEmpire().policyAdopted("INDUSTRIALISM") and stability >= get_named_real("PLC_INDUSTRIALISM_MIN_STABILITY"):
        return 1.0 + get_named_real("PLC_INDUSTRIALISM_TARGET_INDUSTRY_PERCENT")
    return 1.0


def _get_production_bonus_mod_by_policy(stability: float) -> float:
    """
    Calculate bonus production per population which we would get independent of the species production skill,
    but still affected by industrialism.
    """
    # TBD: check connections?
    bonuses = [
        Bonus(
            bool(BuildingType.BLACK_HOLE_POW_GEN.built_or_queued_at()),
            get_named_real("BLD_BLACK_HOLE_POW_GEN_MIN_STABILITY"),
            get_named_real("BLD_BLACK_HOLE_POW_GEN_TARGET_INDUSTRY_PERPOP"),
        ),
    ]
    return sum(bonus.get_bonus(stability) for bonus in bonuses)


def _get_production_bonus_unmodified(planet: fo.planet, stability: float) -> float:
    """
    Calculate bonus production per population which we would get independent of the species production skill.
    """
    specials_bonus = sum(
        # growth.macros: STANDARD_INDUSTRY_BOOST
        AIDependencies.INDUSTRY_PER_POP
        for s in planet.specials
        if s in AIDependencies.industry_boost_specials_unmodified
    )

    # TBD: check connections?
    bonuses = [
        Bonus(
            bool(BuildingType.SOL_ORB_GEN.built_or_queued_at()),  # TBD: check star type
            get_named_real("BLD_SOL_ORB_GEN_MIN_STABILITY"),
            get_named_real("BLD_SOL_ORB_GEN_BRIGHT_TARGET_INDUSTRY_PERPOP"),
        ),
        Bonus(
            have_honeycomb(),
            0.0,
            get_named_real("HONEYCOMB_TARGET_INDUSTRY_PERPOP"),
        ),
        # TBD: Collective Thought Network? Ignored for now, AI doesn't know how to handle it.
        # It shouldn't make a big difference for selecting which planets to colonise anyway.
    ]
    return specials_bonus + sum(bonus.get_bonus(stability) for bonus in bonuses)


def _get_production_flat(planet: fo.planet, stability: float) -> float:
    """
    Calculate population independent production bonus.
    """
    value = _get_asteroid_and_ggg_value(planet, stability)
    # this is a rather expensive one, so consider it only if it is on top of the list
    if tech_soon_available(AIDependencies.PRO_AUTO_2, 1):
        value += get_named_real("PRO_SENTIENT_AUTO_TARGET_INDUSTRY_FLAT")
    return value


def _get_asteroid_and_ggg_value(planet: fo.planet, stability: float) -> float:
    """
    Calculate an estimate of the bonus we may get from asteroids (microgravity) and gas giant generators.
    """
    universe = fo.getUniverse()
    system = universe.getSystem(planet.systemID)
    ast_min_stability = get_named_real("PRO_MICROGRAV_MAN_MIN_STABILITY")
    count_asteroids = ast_min_stability <= stability and tech_soon_available(AIDependencies.PRO_MICROGRAV_MAN, 5)
    ggg_min_stability = get_named_real("BLD_GAS_GIANT_GEN_MIN_STABILITY")
    count_ggg = ggg_min_stability <= stability and tech_soon_available(AIDependencies.PRO_ORBITAL_GEN, 5)
    asteroid_value = 0.0
    ggg_value = 0.0
    for pid in system.planetIDs:
        p2 = universe.getPlanet(pid)
        if p2:
            if count_asteroids and p2.size == fo.planetSize.asteroids:
                full_bonus = get_named_real("PRO_MICROGRAV_MAN_TARGET_INDUSTRY_FLAT")
                if p2.owner == fo.empireID or pid == planet.id:
                    asteroid_value = full_bonus
                # If we cannot see it, we cannot put an outpost on it, although in case of asteroids it likely
                # only hidden temporarily.
                elif universe.getVisibility(pid, fo.empireID()) and p2.unowned:
                    asteroid_value = max(asteroid_value, 0.5 * full_bonus)
                # TODO prospective_invasion_targets?
            if count_ggg:
                ggg_value = max(ggg_value, _ggg_value(planet, p2))
    debug_rating(
        f"_get_asteroid_and_ggg_value amin={ast_min_stability:.1f}->{count_asteroids:.1f}, "
        f"gmin={ggg_min_stability:.1f}->{count_ggg:.1f}, aval={asteroid_value:.1f}, gval={ggg_value:.1f}"
    )
    return asteroid_value + ggg_value


def _ggg_value(planet: fo.planet, p2: fo.planet) -> float:
    """
    Check if p2 not planet and is a gas giant. If determines a value for the potential GGG on p2.
    """
    universe = fo.getUniverse()
    if p2.id != planet.id and p2.size == fo.planetSize.gasGiant:
        ggg_colony_flat = get_named_real("BLD_GAS_GIANT_GEN_COLONY_TARGET_INDUSTRY_FLAT")
        ggg_outpost_flat = get_named_real("BLD_GAS_GIANT_GEN_OUTPOST_TARGET_INDUSTRY_FLAT")
        # species living on a GG do not like a GGG, so we ignore the planet itself here
        if p2.owner == fo.empireID():
            if not p2.speciesName:
                return ggg_outpost_flat
            if p2.id in BuildingType.GAS_GIANT_GEN.built_or_queued_at():
                return ggg_colony_flat
            else:
                return 0.4 * ggg_colony_flat  # we could build one, but inhabitants won't like it
        # If we cannot see it, we cannot put an outpost on it, most likely it is not really unowned
        if universe.getVisibility(p2.id, fo.empireID()) and p2.unowned:
            return 0.5 * ggg_colony_flat  # unowned means we could probably get it easily
        # TODO prospective_invasion_targets?
        #  So far we do not consider owned gas giant for planets here
    return 0.0



from collections.abc import Iterable
from dataclasses import dataclass
from typing import TextIO


@dataclass
class Import:
    from_: str
    items: list[str]


def merge_imports(imports):
    import_map = {}
    for i in imports:
        if i.from_ in import_map:
            import_map[i.from_].items.extend(i.items)
        else:
            import_map[i.from_] = i
    return list(import_map.values())


def sort_imports(imports):
    imports.sort(key=lambda x: x.from_)
    for import_ in imports:
        import_.items.sort()


class ResultBuilder:
    def __init__(self, docstring: str):
        self._docstring = docstring
        self._built_in_imports = []
        self._imports = []
        self._extra_declaration = []
        self._classes = []
        self._enums = []
        self._functions = []

    def add_built_in_import(self, import_: Import):
        self._built_in_imports.append(import_)

    def add_import(self, import_: Import):
        self._imports.append(import_)

    def add_extra_declaration(self, declaration: str):
        self._extra_declaration.append(declaration)

    def add_classes(self, classes: Iterable[str]):
        self._classes.extend(classes)

    def add_enums(self, enums: Iterable[str]):
        self._enums.extend(enums)

    def add_functions(self, functions: Iterable[str]):
        self._functions.extend(functions)

    def _import_string(self, import_: Import):
        return f"from {import_.from_} import {', '.join(import_.items)}"

    def _write_imports(self, file: TextIO, imports: list[Import]):
        imports = merge_imports(imports)
        sort_imports(imports)
        for i in imports:
            oneliner = self._import_string(i)
            if len(oneliner) <= 100:
                file.write(oneliner)
                file.write("\n")
            else:
                file.write(f"from {i.from_} import (\n")
                for item in i.items:
                    file.write(f"    {item},\n")
                file.write(")\n")

    def write(self, file: TextIO):
        file.write(self._docstring)
        file.write("\n")
        self._write_imports(file, self._built_in_imports)
        file.write("\n")
        self._write_imports(file, self._imports)

        for declaraion in self._extra_declaration:
            file.write(declaraion)
            file.write("\n")

        for cls in self._classes:
            file.write("\n")
            file.write(cls)
            file.write("\n")

        for enum in self._enums:
            file.write("\n")
            file.write(enum)
            file.write("\n")

        file.write("\n")
        for function in self._functions[:-1]:
            file.write(function)
            file.write("\n")

        last_function = self._functions[-1]
        file.write(last_function)
        if not last_function.endswith("\n"):
            file.write("\n")



# Module for typing checks in sceleton files
# It should be real py file to be importable, but this file is not actually used in runtime.
#
# Game still try to import it and find the FOCS objects, so we use try catch to supress an error in logs.
# Python do not have typing library included in FOCS parsing environment.
# See discussion here: https://github.com/freeorion/freeorion/issues/4844

# mypy: disable-error-code="empty-body"
try:
    from typing import TypeVar, Union

    from typing_extensions import Self

    class _MeterType:
        ...

    class _Effect:
        ...

    class _EffectGroup:
        ...

    class _StarType:
        ...

    class _Visibility:
        ...

    class _PlanetSize:
        ...

    class _PlanetType:
        ...

    class _PlanetEnvironment:
        ...

    # Type for return
    class _Empire:
        ...

    # Type for arguments, since invalid Empire is -1
    _EmpireId = Union[_Empire, int]

    class _Focus:
        ...

    class _Resource:
        ...

    class _Aggregator:
        ...

    _T = TypeVar("_T", str, int, float)

    class _Condition(bool):  # type: ignore
        """
        This value represent an expression that is resolved to boolean.

        We need to have a special "class" because we use integer operands with it (& | ~),
        which return int type, that breaks linter logic.
        """

        def __and__(self, other) -> Self:
            ...

        def __or__(self, other) -> Self:
            ...

        def __invert__(self) -> Self:
            ...

    class _ConditionalComposition:
        """
        Base class to type that return conditions when used in comparison operations.
        """

        def __eq__(self, other) -> _Condition:  # type: ignore[override]
            ...

        def __ne__(self, other) -> _Condition:  # type: ignore[override]
            ...

        def __lt__(self, other) -> _Condition:
            ...

        def __gt__(self, other) -> _Condition:
            ...

        def __le__(self, other) -> _Condition:
            ...

        def __ge__(self, other) -> _Condition:
            ...

    class _ID:
        ...

        def __eq__(self, other) -> _Condition:
            ...

        def __ne__(self, other) -> _Condition:
            ...

    class _SystemID(_ID):
        ...

    class _PlanetId(_ID):
        ...

    class _DesignID(_ID):
        ...

    _SpeciesValue = str

    # Type hints for function arguments.  This could be a literal or _Value returned by other function.
    _StringParam = str
    _IntParam = int
    # We assume that we accept int anywhere we could accept float
    _FloatParam = Union[int, float]
    _ValueParam = Union[int, float, str]

    class _BuildingType:
        ...

except ModuleNotFoundError:
    pass



class ShellVariable:
    """
    Variable initialization and description for an interpreter.
    """

    def __init__(self, *, variable: str, expression: str, description: str, imports=tuple()):
        self.name = variable
        self.expression = expression
        self.description = description
        self.imports = imports

    def get_evaluation_command(self) -> str:
        yield from self.imports
        yield f"{self.name} = {self.expression}"



from focs._effects import (
    Adequate,
    EffectsGroup,
    Good,
    OwnedBy,
    OwnerHasTech,
    Planet,
    SetTargetPopulation,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_POPULATION_AFTER_SCALING_PRIORITY

Tech(
    name="GRO_PLANET_ECOL",
    description="GRO_PLANET_ECOL_DESC",
    short_description="POPULATION_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=8 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_GROWTH_CATEGORY"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & ~OwnerHasTech(name="GRO_SYMBIOTIC_BIO")
            & OwnedBy(empire=Source.Owner)
            & Planet(environment=[Good, Adequate]),
            accountinglabel="GRO_TECH_ACCOUNTING_LABEL",
            priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
            effects=SetTargetPopulation(value=Value + 1),
        )
    ],
    graphic="icons/tech/planetary_ecology.png",
)



import pytest

from freeorion_tools import combine_ratings


@pytest.mark.parametrize(
    ("rating1", "raring2", "combined"),
    (
        (1, 1, pytest.approx(4.0)),
        (2, 1, pytest.approx(5.82, rel=1e-2)),
        (2, 2, pytest.approx(8.0)),
        (4, 2, pytest.approx(11.65, rel=1e-2)),
    ),
)
def test_two_rating_are_merged(rating1, raring2, combined):
    assert combine_ratings(rating1, raring2) == combined


def test_merge_3_ratings():
    assert combine_ratings(1, 1, 2) == pytest.approx(11.65, rel=1e-2)


@pytest.mark.parametrize(
    ("iterable"),
    (
        (1, 1, 2),
        iter((1, 1, 2)),
        [1, 1, 2],
    ),
)
def test_merge_3_ratings_as_collection(iterable):
    assert combine_ratings(iterable) == pytest.approx(11.65, rel=1e-2)


def test_merge_3_ratings_as_collection_and_args():
    assert combine_ratings([1, 1], 2) == pytest.approx(11.65, rel=1e-2)


def test_idempotency():
    assert combine_ratings(1, 1, 2) == combine_ratings(2, 1, 1) == combine_ratings(1, 2, 1)



import freeorion as fo
import random
import sys
from itertools import product

import planets
import universe_tables
import util

# tuple of available star types
star_types = (
    fo.starType.blue,
    fo.starType.white,
    fo.starType.yellow,
    fo.starType.orange,
    fo.starType.red,
    fo.starType.neutron,
    fo.starType.blackHole,
    fo.starType.noStar,
)

# tuple of "real" star types (that is, blue to red, not neutron, black hole or even no star)
star_types_real = (fo.starType.blue, fo.starType.white, fo.starType.yellow, fo.starType.orange, fo.starType.red)


def pick_star_type(galaxy_age):
    """
    Picks and returns a star type based on universe tables distribution modifiers.
    """

    # try to pick a star type by making a series of "rolls" (1-100)
    # for each star type, and take the highest modified roll
    star_type = fo.starType.unknown
    try:
        max_roll = 0
        for candidate in star_types:
            roll = (
                random.randint(1, 100)
                + universe_tables.UNIVERSE_AGE_MOD_TO_STAR_TYPE_DIST[galaxy_age][candidate]
                + universe_tables.BASE_STAR_TYPE_DIST[candidate]
            )
            if max_roll < roll:
                max_roll = roll
                star_type = candidate
    except:  # noqa: E722
        # in case of an error play save and set star type to invalid
        star_type = fo.starType.unknown
        util.report_error("Python pick_star_type: Pick star type failed\n" + sys.exc_info()[1])

    # if we got an invalid star type (for whatever reason),
    # just select one randomly from the global tuple
    if star_type == fo.starType.unknown:
        star_type = random.choice(star_types)
    return star_type


def name_planets(system):
    """
    Sets the names of the planets of the specified system.

    Planet name is system name + planet number (as roman number)
    unless it's an asteroid belt, in that case name is system
    name + 'asteroid belt' (localized).
    """
    # iterate over all planets in the system
    sys_name = fo.get_name(system)
    for planet in fo.sys_get_planets(system):
        name = fo.userString("NEW_PLANET_NAME")
        name = name.replace("%1%", sys_name)
        name = name.replace("%2%", fo.planet_cardinal_suffix(planet))
        name = name.strip()
        fo.set_name(planet, name)


def can_have_no_planets(star_type):
    """
    Return True is system is allowed to have no planets.
    This intentionally only checks the STAR_TYPE_MOD_TO_PLANET_SIZE_DIST so that
    the following code forces most stars to have a planetary body so that in the
    GUI the presence of a star implies a planet
    """
    return universe_tables.STAR_TYPE_MOD_TO_PLANET_SIZE_DIST[star_type][fo.planetSize.noWorld] > 0


def generate_systems(pos_list, gsd):
    """
    Generates and populates star systems at all positions in specified list.
    """
    sys_list = []
    for position in pos_list:
        star_type = pick_star_type(gsd.age)
        system = fo.create_system(star_type, "", position[0], position[1])
        if system == fo.invalid_object():
            # create system failed, report an error and try to continue with next position
            util.report_error(
                f"Python generate_systems: create system at position ({position[0]:f}, {position[1]:f}) failed"
            )
            continue
        sys_list.append(system)

        orbits = list(range(fo.sys_get_num_orbits(system)))

        if not planets.can_have_planets(star_type, orbits, gsd.planet_density, gsd.shape):
            continue

        # Try to generate planets in each orbit.
        # If after each orbit is tried once there are no planets then
        # keep trying until a single planet is placed.
        # Except for black hole systems, which can be empty.

        at_least_one_planet = False
        random.shuffle(orbits)
        for orbit in orbits:
            if planets.generate_a_planet(system, star_type, orbit, gsd.planet_density, gsd.shape):
                at_least_one_planet = True

        if at_least_one_planet or can_have_no_planets(star_type):
            continue

        recursion_limit = 1000
        for _, orbit in product(range(recursion_limit), orbits):
            if planets.generate_a_planet(system, star_type, orbit, gsd.planet_density, gsd.shape):
                break
        else:
            # Intentionally non-modal.  Should be a warning.
            print(
                (
                    "Python generate_systems: place planets in system %d at position (%.2f, %.2f) failed"
                    % (system, position[0], position[1])
                ),
                file=sys.stderr,
            )

    return sys_list



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="PRO_NANOTECH_PROD",
    description="PRO_NANOTECH_PROD_DESC",
    short_description="THEORY_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=80 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_PRODUCTION_CATEGORY", "THEORY"],
    prerequisites=["PRO_ROBOTIC_PROD"],
    graphic="icons/tech/nanotech_production.png",
)



from focs._effects import EffectsGroup, HasTag, IsSource, OnPlanet, Planet, SetStealth, Source, Value
from macros.stealth import HIGH_STEALTH, LOW_STEALTH, MEDIUM_STEALTH

# Linter get crazy because of common suffix for two different files common/stealth.py
assert isinstance(LOW_STEALTH, int)
assert isinstance(MEDIUM_STEALTH, int)
assert isinstance(HIGH_STEALTH, int)

BAD_STEALTH = [
    EffectsGroup(
        description="BAD_STEALTH_DESC",
        scope=IsSource | (OnPlanet(id=Source.ID) & ~HasTag(name="ORBITAL")),
        activation=Planet(),
        stackinggroup="SPECIES_STEALTH_STACK",
        effects=SetStealth(value=Value - LOW_STEALTH),
    )
]

GOOD_STEALTH = [
    EffectsGroup(
        description="GOOD_STEALTH_DESC",
        scope=IsSource | (OnPlanet(id=Source.ID) & ~HasTag(name="ORBITAL")),
        activation=Planet(),
        stackinggroup="SPECIES_STEALTH_STACK",
        effects=SetStealth(value=Value + LOW_STEALTH),
    )
]


GREAT_STEALTH = [
    EffectsGroup(
        description="GREAT_STEALTH_DESC",
        scope=IsSource | (OnPlanet(id=Source.ID) & ~HasTag(name="ORBITAL")),
        activation=Planet(),
        stackinggroup="SPECIES_STEALTH_STACK",
        effects=SetStealth(value=Value + MEDIUM_STEALTH),
    )
]


ULTIMATE_STEALTH = [
    EffectsGroup(
        description="ULTIMATE_STEALTH_DESC",
        scope=IsSource | (OnPlanet(id=Source.ID) & ~HasTag(name="ORBITAL")),
        activation=Planet(),
        stackinggroup="SPECIES_STEALTH_STACK",
        effects=SetStealth(value=Value + HIGH_STEALTH),
    )
]



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_COLLAPSER",
    description="LRN_COLLAPSER_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=1440 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_ART_BLACK_HOLE", "LRN_STELLAR_TOMOGRAPHY", "LRN_SPATIAL_DISTORT_GEN"],
    unlock=Item(type=UnlockBuilding, name="BLD_BLACK_HOLE_COLLAPSER"),
    graphic="icons/tech/black_hole_collapse.png",
)



from focs._effects import (
    Abs,
    AnyEmpire,
    BlackHole,
    Blue,
    Contains,
    CurrentTurn,
    EffectsGroup,
    Focus,
    GasGiantType,
    Good,
    HasSpecies,
    HasTag,
    Homeworld,
    Hostile,
    IsSource,
    LocalCandidate,
    MaxOf,
    MinOf,
    Neutron,
    NoStar,
    Orange,
    OwnedBy,
    Planet,
    Poor,
    Red,
    ResourceSupplyConnected,
    SetPopulation,
    SetStarType,
    SetTargetPopulation,
    Source,
    Star,
    Target,
    TargetPopulation,
    Turn,
    Uninhabitable,
    Value,
    White,
    Yellow,
)
from macros.misc import GROWTH_RATE_FACTOR
from macros.priorities import (
    POPULATION_FIRST_PRIORITY,
    TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
    TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
    TARGET_POPULATION_OVERRIDE_PRIORITY,
    TARGET_POPULATION_SCALING_PRIORITY,
)
from species.species_macros.advanced_focus import ADVANCED_FOCUS_EFFECTS
from species.species_macros.general import (
    FOCUS_CHANGE_PENALTY,
    STANDARD_CONSTRUCTION,
    STANDARD_METER_GROWTH,
)

_HOMEWORLD_BONUS_POPULATION = EffectsGroup(
    scope=IsSource & Homeworld(name=[Source.Species]),
    activation=Planet(),
    stackinggroup="HOMEWORLD_STACK",
    accountinglabel="HOMEWORLD_BONUS",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 2 * Target.HabitableSize),
)


# Implements environmental modifiers from Growth techs.
# Changes to the growth tree should take this into account.

_ENVIRONMENT_MODIFIER = [
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Uninhabitable]),
        accountinglabel="UNINHABTIABLE_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=-999),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Hostile]),
        accountinglabel="HOSTILE_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 4 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Planet(environment=[Poor]),
        accountinglabel="POOR_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 2 * Source.HabitableSize),
    ),
    #           EffectsGroup
    #             scope = Source
    #             activation = And [
    #                 Planet
    #                 Planet environment = Adequate
    #             ]
    #             accountinglabel = "ADEQUATE_ENVIRONMENT_LABEL"
    #             priority = [[TARGET_POPULATION_BEFORE_SCALING_PRIORITY]]
    #             effects = SetTargetPopulation value = Value + 0 * Source.HabitableSize
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Good]),
        accountinglabel="GOOD_ENVIRONMENT_LABEL",
        priority=TARGET_POPULATION_BEFORE_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value + 3 * Target.HabitableSize),
    ),
]

# This is dependent on current placement in population effects calc,
# just after Homeworld and Environment
_SELF_SUSTAINING_BONUS = EffectsGroup(
    scope=IsSource,
    activation=Planet(environment=[Good]) & HasTag(name="SELF_SUSTAINING"),
    accountinglabel="SELF_SUSTAINING_LABEL",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 3 * Target.HabitableSize),  # Gets the same bonus as three growth specials
)

_PHOTOTROPHIC_BONUS = [
    EffectsGroup(
        scope=Contains(IsSource),
        activation=Planet()
        & (
            OwnedBy(affiliation=AnyEmpire)
            | Star(type=[BlackHole, NoStar])  # Natives are allowed to have bright or dim stars
        )
        & HasTag(name="PHOTOTROPHIC")
        & Turn(high=0)
        & ~Star(type=[Orange]),
        effects=SetStarType(type=Yellow),  # start with a normal star to be balanced
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[Blue]) & TargetPopulation(low=0),
        accountinglabel="VERY_BRIGHT_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value + 3 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[White]) & TargetPopulation(low=0),
        accountinglabel="BRIGHT_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value + 1.5 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[Red, Neutron]),
        accountinglabel="DIM_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value - 1 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & HasTag(name="PHOTOTROPHIC") & Star(type=[BlackHole, NoStar]),
        accountinglabel="NO_STAR",
        priority=TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY,
        effects=SetTargetPopulation(value=Value - 10 * Source.HabitableSize),
    ),
]

_GASEOUS_BONUS = EffectsGroup(
    scope=IsSource,
    activation=Planet(type=[GasGiantType]) & HasTag(name="GASEOUS"),
    accountinglabel="GASEOUS_LABEL",
    priority=TARGET_POPULATION_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value - 0.5 * Abs(float, Value)),
)

_HOMEWORLD_GROWTH_FOCUS_BOOST = EffectsGroup(
    scope=Planet()
    & OwnedBy(empire=Source.Owner)
    & HasSpecies(name=[Source.Species])
    & ~Homeworld(name=[Source.Species])
    & ResourceSupplyConnected(empire=Source.Owner, condition=IsSource),
    activation=Planet() & Focus(type=["FOCUS_GROWTH"]) & Homeworld(),
    stackinggroup="HOMEWORLD_STACK",
    accountinglabel="HOMEWORLD_SUPPLY",
    priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
    effects=SetTargetPopulation(value=Value + 1 * Target.HabitableSize),
)


_BASIC_POPULATION = [
    _HOMEWORLD_BONUS_POPULATION,
    *_ENVIRONMENT_MODIFIER,
    _SELF_SUSTAINING_BONUS,
    *_PHOTOTROPHIC_BONUS,
    _GASEOUS_BONUS,
    _HOMEWORLD_GROWTH_FOCUS_BOOST,
    # population growth or decay towards to target population
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & (LocalCandidate.LastTurnConquered < CurrentTurn),
        priority=POPULATION_FIRST_PRIORITY,
        effects=SetPopulation(
            # growth is GROWTH_RATE_FACTOR * population * (maximum + 1 - population) / maximum
            # which gives factor for population 1, best growth around half full and aproaching
            # factor again when almost full
            value=(
                (Value < Value(Target.TargetPopulation))
                * MinOf(
                    float,
                    Value(Target.TargetPopulation),
                    Value + Value * GROWTH_RATE_FACTOR * (1 + (1 - Value) / Value(Target.TargetPopulation)),
                )
            )
            # shrink if overpopulated
            + (
                (Value >= Value(Target.TargetPopulation))
                * MaxOf(float, Value(Target.TargetPopulation), Value - 0.1 * (Value - Value(Target.TargetPopulation)))
            )
        ),
    ),
    # Since all species have the same advanced focus effects and
    # infrastructure, the macros are stashed here, where they don't
    # need to be manually included in each species' macros.
    *FOCUS_CHANGE_PENALTY,
    *ADVANCED_FOCUS_EFFECTS,
    *STANDARD_CONSTRUCTION,
    *STANDARD_METER_GROWTH,
]

EXTREMELY_BAD_POPULATION = [
    *_BASIC_POPULATION,
    EffectsGroup(
        description="EXTREMELY_BAD_POPULATION_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="EXTREMELY_BAD_POPULATION_LABEL",
        priority=TARGET_POPULATION_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 0.75 * Abs(float, Value)),
    ),
]


VERY_BAD_POPULATION = [
    *_BASIC_POPULATION,
    EffectsGroup(
        description="VERY_BAD_POPULATION_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="VERY_BAD_POPULATION_LABEL",
        priority=TARGET_POPULATION_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 0.5 * Abs(float, Value)),
    ),
]

BAD_POPULATION = [
    *_BASIC_POPULATION,
    EffectsGroup(
        description="BAD_POPULATION_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="BAD_POPULATION_LABEL",
        priority=TARGET_POPULATION_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 0.25 * Abs(float, Value)),
    ),
]

AVERAGE_POPULATION = _BASIC_POPULATION

GOOD_POPULATION = [
    *_BASIC_POPULATION,
    EffectsGroup(
        description="GOOD_POPULATION_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GOOD_POPULATION_LABEL",
        priority=TARGET_POPULATION_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value + 0.25 * Abs(float, Value)),
    ),
]

GREAT_POPULATION = [
    *_BASIC_POPULATION,
    EffectsGroup(
        description="GREAT_POPULATION_DESC",
        scope=IsSource,
        activation=Planet(),
        accountinglabel="GREAT_POPULATION_LABEL",
        priority=TARGET_POPULATION_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value + 0.5 * Abs(float, Value)),
    ),
]


# reduces max population in systems with bright stars
LIGHT_SENSITIVE = [
    EffectsGroup(
        description="LIGHT_SENSITIVE_DESC",
        scope=IsSource,
        activation=Planet() & Star(type=[Blue]),
        accountinglabel="VERY_BRIGHT_STAR",
        priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - 2 * Source.HabitableSize),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Star(type=[White]),
        accountinglabel="BRIGHT_STAR",
        priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
        effects=SetTargetPopulation(value=Value - Source.HabitableSize),
    ),
]



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SINGULARITY_ENGINE_CORE",
    description="SHP_SINGULARITY_ENGINE_CORE_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=800 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_ENGINE_PART_TECHS"],
    prerequisites=["SHP_N_DIMENSIONAL_ENGINE_MATRIX", "PRO_SINGULAR_GEN"],
    unlock=Item(type=UnlockShipPart, name="FU_SINGULARITY_ENGINE_CORE"),
    graphic="icons/ship_parts/engine-3.png",
)



from freeorion_tools.chat_handler.base_chat_handler import ChatHandlerBase


class NormalChatHandler(ChatHandlerBase):
    """
    Class for normal message handling.

    It just does nothing and allows next handler to handle that message.
    """

    def process_message(self, sender_id: int, message: str) -> bool:
        return False



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_ALGO_ELEGANCE",
    description="LRN_ALGO_ELEGANCE_DESC",
    short_description="RESEARCH_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=18 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_LEARNING_CATEGORY"],
    unlock=Item(type=UnlockPolicy, name="PLC_ALGORITHMIC_RESEARCH"),
    graphic="icons/tech/algorithmic_elegance.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import NARROW_EP, OCEAN_NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS

Species(
    name="SP_SSLITH",
    description="SP_SSLITH_DESC",
    gameplay_description="SP_SSLITH_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "TELEPATHIC",
        "ARTISTIC",
        "BAD_INDUSTRY",
        "AVERAGE_SUPPLY",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "ELERIUM_SPECIAL",
        "FOCUS_RESEARCH",
        "FRACTAL_GEODES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "MONOPOLE_SPECIAL",
        "PLC_CONFEDERATION",
        "PLC_DREAM_RECURSION",
        "PLC_LIBERTY",
        "PLC_MODERATION",
        "SUPERCONDUCTOR_SPECIAL",
        "TEMPORAL_ANOMALY_SPECIAL",
    ],
    dislikes=[
        # "BLD_CLONING_CENTER"       // disabled content,
        "BLD_ENCLAVE_VOID",
        "BLD_GATEWAY_VOID",
        "BLD_MILITARY_COMMAND",
        "BLD_SHIPYARD_AST",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SPATIAL_DISTORT_GEN",
        "BLD_STARGATE",
        "GAIA_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PLC_CHECKPOINTS",
        "PLC_CONFORMANCE",
        "PLC_DIVINE_AUTHORITY",
        "PLC_TRAFFIC_CONTROL",
        "PROBIOTIC_SPECIAL",
        "WORLDTREE_SPECIAL",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_SSLITH"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=OCEAN_NARROW_EP,
    graphic="icons/species/sslith.png",
)



from collections.abc import Mapping, Sequence
from typing import Union

import AIDependencies
from common.fo_typing import PlanetId, SpeciesName
from empire.survey_lock import survey_universe_lock
from freeorion_tools import tech_is_complete, tech_soon_available
from freeorion_tools.caching import cache_for_current_turn


def set_colony_builders(species_name: SpeciesName, yards: Sequence[PlanetId]):
    """
    Add planet where you can build colonies for species.

    Warning! Temporal coupling.
    All calls of this function should be done before using of this information.
    """
    empire_colonizers = _get_colony_builders()
    empire_colonizers.setdefault(species_name, []).extend(yards)


@survey_universe_lock
def can_build_colony_for_species(species_name: Union[SpeciesName, str]):
    return species_name in get_colony_builders()


@survey_universe_lock
def get_colony_builder_locations(species_name: SpeciesName) -> list[PlanetId]:
    return get_colony_builders()[species_name]


@survey_universe_lock
def can_build_only_sly_colonies():
    """
    Return true if empire could build only SP_SLY colonies.

    This could be possible only on early stage, when no other races are conquered
    and techs like EXOBOTS are not invented.

    This race has poor supply and live only on gas giants.
    """
    return list(get_colony_builders()) == ["SP_SLY"]


@survey_universe_lock
def get_colony_builders() -> Mapping[SpeciesName, list[PlanetId]]:
    """
    Return map from the species to list of the planet where you could build a colony ship with it.
    """
    return _get_colony_builders()


@cache_for_current_turn
def get_extra_colony_builders() -> list[str]:
    """
    Returns species the empire can build without having a colony, i.e. Exobots, if (almost) researched, plus
    extinct species that has been enabled.
    """
    ret = []
    if tech_soon_available(AIDependencies.EXOBOT_TECH_NAME, 1):
        ret.append("SP_EXOBOT")
    for spec_name in AIDependencies.EXTINCT_SPECIES:
        if tech_is_complete("TECH_COL_" + spec_name):
            ret.append("SP_" + spec_name)
    return ret


@cache_for_current_turn
def _get_colony_builders() -> dict[SpeciesName, list[PlanetId]]:
    """
    Return mutable state.
    """
    colony_build_locations = {}

    # get it into colonizer list even if no colony yet
    for species in get_extra_colony_builders():
        colony_build_locations[species] = []
    return colony_build_locations



from focs._effects import (
    Contains,
    DirectDistanceBetween,
    EffectsGroup,
    Focus,
    Happiness,
    IsSource,
    LocalCandidate,
    Max,
    MaxOf,
    MinOf,
    NamedReal,
    OwnedBy,
    Planet,
    SetEmpireMeter,
    SetTargetResearch,
    Source,
    Statistic,
    System,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

Tech(
    name="LRN_DISTRIB_THOUGHT",
    description="LRN_DISTRIB_THOUGHT_DESC",
    short_description="RESEARCH_AND_SLOT_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_PSIONICS"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=NamedReal(name="LRN_DISTRIB_THOUGHT_MIN_STABILITY", value=10)),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + MaxOf(
                    float,
                    0.0,
                    MinOf(
                        float,
                        Target.Construction * NamedReal(name="DISTRIB_THOUGH_INFRA_LIMIT_SCALING", value=0.5),
                        NamedReal(name="DISTRIB_THOUGH_RESEARCH_SCALING", value=0.1)
                        * Statistic(
                            float,
                            Max,
                            value=DirectDistanceBetween(Target.ID, LocalCandidate.ID),
                            condition=System
                            & Contains(Planet() & OwnedBy(empire=Source.Owner) & Focus(type=["FOCUS_RESEARCH"])),
                        )
                        ** 0.5,
                    ),
                )
            ),
        ),
        EffectsGroup(
            scope=IsSource,
            effects=SetEmpireMeter(empire=Source.Owner, meter="MILITARY_CATEGORY_NUM_POLICY_SLOTS", value=Value + 1),
        ),
    ],
    graphic="icons/tech/distributed_thought.png",
)



import freeOrionAIInterface as fo
from collections import defaultdict
from collections.abc import Mapping

import AIstate
from common.fo_typing import SystemId
from freeorion_tools.caching import cache_for_current_turn


def has_claimed_star(*stars: "fo.starType") -> bool:
    """
    Return `True` if at least one star of that type is claimed.
    """
    return bool(set(_get_claimed_stars()).intersection(stars))


def is_system_star_claimed(system: "fo.system"):
    return system.systemID in _get_claimed_stars()[system.starType]


def count_claimed_stars(star_type: "fo.starType") -> int:
    """
    Count claimed starts of specified type.
    """
    return len(_get_claimed_stars()[star_type])


@cache_for_current_turn
def _get_claimed_stars() -> Mapping["fo.starType", set[SystemId]]:
    """
    Return dictionary of star type: list of colonised and planned to be colonized systems.
    Start type converted to int because `cache_by_turn` store its value in savegame
    and boost objects are not serializable.
    """
    claimed_stars = defaultdict(set)

    claimed_stars.update({int(s_type): set(AIstate.empireStars[s_type]) for s_type in AIstate.empireStars})

    universe = fo.getUniverse()
    for sys_id in set(AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs):
        t_sys = universe.getSystem(sys_id)
        if not t_sys:
            continue
        claimed_stars[t_sys.starType].add(sys_id)
    return claimed_stars



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_ORG_HULL",
    description="SHP_ORG_HULL_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=16 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_ORGANIC_HULL_TECHS"],
    prerequisites=["GRO_MEGA_ECO"],
    unlock=[Item(type=UnlockShipHull, name="SH_ORGANIC"), Item(type=UnlockBuilding, name="BLD_SHIPYARD_ORG_ORB_INC")],
    graphic="hulls_design/organic_hull.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_weapons.ship_weapons import WEAPON_UPGRADE_CAPACITY_EFFECTS, WEAPON_UPGRADE_SECONDARY_STAT_EFFECTS

Tech(
    name="SHP_WEAPON_4_3",
    description="SHP_WEAPON_4_3_DESC",
    short_description="SHIP_WEAPON_IMPROVE_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=900 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_WEAPON_4_2"],
    effectsgroups=[
        *WEAPON_UPGRADE_CAPACITY_EFFECTS("SHP_WEAPON_4_3", "SR_WEAPON_4_1", 5),
        *WEAPON_UPGRADE_SECONDARY_STAT_EFFECTS("SHP_WEAPON_4_3", "SR_WEAPON_0_1", 1),
    ],
    graphic="icons/ship_parts/death-ray-3.png",
)



from typing import Callable

from common.statistic_interface import LOG_PREFIX, StatKey, StatValue


class Stats:
    def __init__(self, writer: Callable[[str], None]):
        self._writer = writer

    def write(self, key: StatKey, value: StatValue):
        serialized = key.value.serialize(value)
        self._writer(f"{LOG_PREFIX}{key.value.name}:{serialized}")



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_MASSPROP_SPEC",
    description="SHP_MASSPROP_SPEC_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=1000 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_ROBOTIC_HULL_TECHS"],
    prerequisites=["SHP_CONTGRAV_MAINT"],
    unlock=[Item(type=UnlockShipHull, name="SH_TITANIC")],
    graphic="icons/ship_hulls/titanic_hull_small.png",
)



from focs._effects import (
    EffectsGroup,
    HasSpecies,
    OwnedBy,
    SetTargetPopulation,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_POPULATION_AFTER_SCALING_PRIORITY

Tech(
    name="CON_ORBITAL_HAB",
    description="CON_ORBITAL_HAB_DESC",
    short_description="POPULATION_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=350 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["PRO_MICROGRAV_MAN"],
    effectsgroups=[
        EffectsGroup(
            scope=HasSpecies() & OwnedBy(empire=Source.Owner),
            priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
            effects=SetTargetPopulation(value=Value + 1 * Target.HabitableSize, accountinglabel="ORBITAL_HAB_LABEL"),
        )
    ],
    graphic="icons/tech/orbital_gardens.png",
)



import freeOrionAIInterface as fo
from logging import debug

import AIstate
import EspionageAI
import FleetUtilsAI
import InvasionAI
import PlanetUtilsAI
import PriorityAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from CombatRatingsAI import (
    get_fleet_rating,
    get_fleet_rating_against_planets,
    get_ship_combat_stats,
    rating_difference,
    rating_needed,
)
from EnumsAI import MissionType
from expansion_plans import get_colonisable_outpost_ids, get_colonisable_planet_ids
from freeorion_tools import combine_ratings
from freeorion_tools.caching import cache_by_turn_persistent
from target import TargetSystem
from turn_state import (
    get_distance_to_enemy_supply,
    get_owned_planets,
    get_owned_planets_in_system,
    get_systems_by_supply_tier,
)
from turn_state.design import cur_best_military_design_rating
from universe.system_network import systems_connected

MinThreat = 10  # the minimum threat level that will be ascribed to an unknown threat capable of killing scouts
_military_allocations = []
_verbose_mil_reporting = False
_best_ship_rating_cache = {}  # indexed by turn, value is rating of that turn


def cur_best_mil_ship_rating(include_designs=False):
    """Find the best military ship we have available in this turn and return its rating.

    :param include_designs: toggles if available designs are considered or only existing ships
    :return: float: rating of the best ship
    """
    current_turn = fo.currentTurn()
    if current_turn in _best_ship_rating_cache:
        best_rating = _best_ship_rating_cache[current_turn]
        if include_designs:
            best_design_rating = cur_best_military_design_rating()
            best_rating = max(best_rating, best_design_rating)
        return best_rating
    best_rating = 0.001
    universe = fo.getUniverse()
    aistate = get_aistate()
    for fleet_id in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY):
        fleet = universe.getFleet(fleet_id)
        for ship_id in fleet.shipIDs:
            ship_rating = get_ship_combat_stats(ship_id).get_rating(enemy_stats=aistate.get_standard_enemy())
            best_rating = max(best_rating, ship_rating)
    _best_ship_rating_cache[current_turn] = best_rating
    if include_designs:
        best_design_rating = cur_best_military_design_rating()
        best_rating = max(best_rating, best_design_rating)
    return max(best_rating, 0.001)


def get_preferred_max_military_portion_for_single_battle() -> float:
    """
    Determine and return the preferred max portion of military to be allocated to a single battle.

    May be used to downgrade various possible actions requiring military support if they would require an excessive
    allocation of military forces.  At the beginning of the game this max portion starts as 1.0, then is slightly
    reduced to account for desire to reserve some defenses for other locations, and then in mid to late game, as the
    size of the military grows, this portion is further reduced to promote pursuit of multiple battlefronts in
    parallel as opposed to single battlefronts against heavily defended positions.

    :return: a number in range (0:1] for preferred max portion of military to be allocated to a single battle
    """
    # TODO: this is a roughcut first pass, needs plenty of refinement
    if fo.currentTurn() < 40:
        return 1.0
    best_ship_equivalents = (get_concentrated_tot_mil_rating() / cur_best_mil_ship_rating()) ** 0.5
    _MAX_SHIPS_BEFORE_PREFERRING_LESS_THAN_FULL_ENGAGEMENT = 3
    if best_ship_equivalents <= _MAX_SHIPS_BEFORE_PREFERRING_LESS_THAN_FULL_ENGAGEMENT:
        return 1.0
    # the below ratio_exponent is still very much a work in progress.  It should probably be somewhere in the range of
    # 0.2 to 0.5.  Values at the larger end will create a smaller expected battle size threshold that would
    # cause the respective opportunity (invasion, colonization) scores to be discounted, thereby more quickly creating
    # pressure for the AI to pursue multiple small/medium resistance fronts rather than pursuing a smaller number fronts
    # facing larger resistance.  The AI will start facing some scoring pressure to not need to throw 100% of its
    # military at a target as soon as its max military rating surpasses the equvalent of
    # _MAX_SHIPS_BEFORE_PREFERRING_LESS_THAN_FULL_ENGAGEMENT of its best ships.  That starts simply as some scoring
    # pressure to be able to hold back some small portion of its ships from the engagement, in order to be able to use
    # them for defense or for other targets.  With an exponent value of 0.25, this would start creating substantial
    # pressure against devoting more than half the military to a single target once the total military is somewhere
    # above 18 best-ship equivalents, and pressure against deovting more than a third once the total is above about 80
    # best-ship equivalents.  With an exponent value of 0.5, those thresholds would be 6 ships and 11 ships.  With the
    # initial value of 0.35, those thresholds are about 10 ships and 25 ships.  Depending on how this return value is
    # used, it should not prevent the more heavily fortified targets (and therefore discounted) from being taken
    # if there are no more remaining easier targets available.
    ratio_exponent = 0.35
    return 1.0 / (best_ship_equivalents + 1 - _MAX_SHIPS_BEFORE_PREFERRING_LESS_THAN_FULL_ENGAGEMENT) ** ratio_exponent


def try_again(mil_fleet_ids, try_reset=False, thisround=""):
    """Clear targets and orders for all specified fleets then call get_military_fleets again."""
    aistate = get_aistate()
    for fid in mil_fleet_ids:
        mission = aistate.get_fleet_mission(fid)
        mission.clear_fleet_orders()
        mission.clear_target()
    get_military_fleets(try_reset=try_reset, thisround=thisround)


def avail_mil_needing_repair(mil_fleet_ids, split_ships=False, on_mission=False, repair_limit=0.70):  # noqa: C901
    """Returns tuple of lists: (ids_needing_repair, ids_not)."""
    fleet_buckets = [[], []]
    universe = fo.getUniverse()
    cutoff = [repair_limit, 0.25][on_mission]
    aistate = get_aistate()
    for fleet_id in mil_fleet_ids:
        fleet = universe.getFleet(fleet_id)
        ship_buckets = [[], []]
        ships_cur_health = [0, 0]
        ships_max_health = [0, 0]
        for ship_id in fleet.shipIDs:
            this_ship = universe.getShip(ship_id)
            cur_struc = this_ship.initialMeterValue(fo.meterType.structure)
            max_struc = this_ship.initialMeterValue(fo.meterType.maxStructure)
            ship_ok = cur_struc >= cutoff * max_struc
            ship_buckets[ship_ok].append(ship_id)
            ships_cur_health[ship_ok] += cur_struc
            ships_max_health[ship_ok] += max_struc
        this_sys_id = fleet.systemID if fleet.nextSystemID == INVALID_ID else fleet.nextSystemID
        fleet_ok = sum(ships_cur_health) >= cutoff * sum(ships_max_health)
        local_status = aistate.systemStatus.get(this_sys_id, {})
        my_local_rating = combine_ratings(
            local_status.get("mydefenses", {}).get("overall", 0), local_status.get("myFleetRating", 0)
        )
        my_local_rating_vs_planets = local_status.get("myFleetRatingVsPlanets", 0)
        combat_trigger = bool(local_status.get("fleetThreat", 0) or local_status.get("monsterThreat", 0))
        if not combat_trigger and local_status.get("planetThreat", 0):
            universe = fo.getUniverse()
            system = universe.getSystem(this_sys_id)
            for planet_id in system.planetIDs:
                planet = universe.getPlanet(planet_id)
                if planet.ownedBy(fo.empireID()):  # TODO: also exclude at-peace planets
                    continue
                if planet.unowned and not EspionageAI.colony_detectable_by_empire(planet_id, empire=fo.empireID()):
                    continue
                if sum(
                    [
                        planet.currentMeterValue(meter_type)
                        for meter_type in [fo.meterType.defense, fo.meterType.shield, fo.meterType.construction]
                    ]
                ):
                    combat_trigger = True
                    break
        needed_here = (
            combat_trigger and local_status.get("totalThreat", 0) > 0
        )  # TODO: assess if remaining other forces are sufficient
        safely_needed = (
            needed_here
            and my_local_rating > local_status.get("totalThreat", 0)
            and my_local_rating_vs_planets > local_status.get("planetThreat", 0)
        )  # TODO: improve both assessment prongs
        if not fleet_ok:
            if safely_needed:
                debug(
                    "Fleet %d at %s needs repair but deemed safely needed to remain for defense"
                    % (fleet_id, universe.getSystem(fleet.systemID))
                )
            else:
                if needed_here:
                    debug(
                        "Fleet %d at %s needed present for combat, but is damaged and deemed unsafe to remain."
                        % (fleet_id, universe.getSystem(fleet.systemID))
                    )
                    debug(
                        "\t my_local_rating: {:.1f} ; threat: {:.1f}".format(
                            my_local_rating, local_status.get("totalThreat", 0)
                        )
                    )
                debug("Selecting fleet %d at %s for repair" % (fleet_id, universe.getSystem(fleet.systemID)))
        fleet_buckets[fleet_ok or bool(safely_needed)].append(fleet_id)
    return fleet_buckets


# TODO Move relevant initialization code from get_military_fleets into this class
class AllocationHelper:
    def __init__(self, already_assigned_rating, already_assigned_rating_vs_planets, available_rating, try_reset):
        """
        :param dict already_assigned_rating:
        :param float available_rating:
        """
        self.try_reset = try_reset
        self.allocations = []
        self.allocation_by_groups = {}

        self.available_rating = available_rating
        self._remaining_rating = available_rating

        self.threat_bias = 0.0
        self.safety_factor = get_aistate().character.military_safety_factor()

        self.already_assigned_rating = dict(already_assigned_rating)
        self.already_assigned_rating_vs_planets = dict(already_assigned_rating_vs_planets)
        # store the number of empires which have supply or have supply within 2 jumps of the system
        self.enemy_supply = {
            sys_id: min(2, len(enemies_nearly_supplying_system(sys_id))) for sys_id in fo.getUniverse().systemIDs
        }

    @property
    def remaining_rating(self):
        return self._remaining_rating

    @remaining_rating.setter
    def remaining_rating(self, value):
        self._remaining_rating = max(0, value)

    def allocate(self, group, sys_id, min_rating, min_rating_vs_planets, take_any, max_rating):
        tup = (sys_id, min_rating, min_rating_vs_planets, take_any, max_rating)
        self.allocations.append(tup)
        self.allocation_by_groups.setdefault(group, []).append(tup)
        if self._remaining_rating <= min_rating:
            self._remaining_rating = 0
        else:
            self._remaining_rating = rating_difference(self._remaining_rating, min_rating)


class Allocator:
    """
    Base class for Military allocation for a single system.

    The Allocator class and its subclasses are used to allocate
    military resources for a single system. First, a minimum
    and a maximum military rating are calculated which are
    required / desired based on e.g. threat in the system.

    An Allocator class defines if military resources should
    be allocated even if the minimum requirements are not met
    or if military resources are only allocated if the threshold
    is passed.

    The information is then passed to an AllocationHelper
    instance. Allocating military resources by an Allocator
    does not necessarily mean that military ships are actually
    assigned to that system. It should be understood as a request
    instead. If allocations of higher priority already use all the
    available military resources, no ships can be sent.


    Public methods:
        :allocate(): Calculate the required/desired military resources
                    for the system and enqueue the allocation info
                    to the AllocationHelper.


    Public attributes:
        :ivar sys_id: ID of the system for which military resources are allocated


    Example usage:
        CapitelDefenseAllocator(capital_sys_id, allocation_helper).allocate()
    """

    _min_alloc_factor = 1.0
    _max_alloc_factor = 2.0
    _potential_threat_factor = 1.0
    _allocation_group = ""
    _military_reset_ratio = -1  # if ratio of available to needed rating is smaller than this, then reset allocations

    def __init__(self, sys_id, allocation_helper):
        """
        :param int sys_id: System for which military resources are allocated
        :param AllocationHelper allocation_helper: The allocation helper where the information is to be stored.
        """
        self.sys_id = sys_id
        self._allocation_helper = allocation_helper

    def allocate(self):
        """Calculate the desired allocation for this system and enqueue it in the allocation_helper."""
        threat = self._calculate_threat()
        min_alloc = self._minimum_allocation(threat)
        max_alloc = self._maximum_allocation(threat)
        alloc_vs_planets = self._allocation_vs_planets()
        if min_alloc <= 0 and alloc_vs_planets <= 0:
            # nothing to allocate here...
            return
        min_alloc = max(min_alloc, alloc_vs_planets)
        max_alloc = max(max_alloc, alloc_vs_planets)

        ratio = self._allocation_helper.remaining_rating / float(min_alloc)
        if self._allocation_helper.remaining_rating > min_alloc or self._take_any():
            self._allocation_helper.allocate(
                group=self._allocation_group,
                sys_id=self.sys_id,
                min_rating=min(min_alloc, self._allocation_helper.remaining_rating),
                min_rating_vs_planets=min(alloc_vs_planets, self._allocation_helper.remaining_rating),
                take_any=self._take_any(),
                max_rating=max_alloc,
            )
        if ratio < 1:
            self._handle_not_enough_resources(ratio)

    def _calculate_threat(self) -> float:
        """
        Calculate the required military rating in the system.

        The value calculated does not have to represent a tangible
        threat / enemy force. It only provides a measurement how much
        military should be sent to the system. Deriving further conditions
        for military presence and translating them into an equivalent
        military rating is strongly encouraged.
        It is implied however, that the value calculated here should
        be greater than or at least equal to the actual visible strength
        of enemy forces within the system so that a subsequent military
        mission can be successful.

        :return: Equivalent military rating required in the system
        """
        raise NotImplementedError

    def _minimum_allocation(self, threat: float) -> float:
        """
        Calculate the minimum allocation for the system.

        The default minimum allocation is the missing forces
        to obtain a rating given by the threat weighted with
        the subclass' *min_alloc_factor*.
        Existing military missions are considered.

        Subclasses may choose to override this method and
        implement a different logic.

        :param threat: threat as calculated by _calculate_threat()
        """
        return rating_needed(self._min_alloc_factor * threat, self.assigned_rating)

    def _maximum_allocation(self, threat: float) -> float:
        """
        Calculate the maximum allocation for the system.

        The default maximum allocation is the missing forces
        to obtain a rating given by the threat weighted with
        the subclass' *max_alloc_factor*.
        Existing military missions are considered.

        Subclasses may choose to override this method and
        implement a different logic.
        """
        return rating_needed(self._max_alloc_factor * threat, self.assigned_rating)

    def _allocation_vs_planets(self):
        return rating_needed(self.safety_factor * self._planet_threat(), self.assigned_rating_vs_planets)

    def _take_any(self) -> bool:
        """
        If true, forces smaller than the minimum allocation are accepted.
        """
        raise NotImplementedError

    def _handle_not_enough_resources(self, ratio: float):
        """Called if minimum allocation is larget than available resources.

        High priority subclasses are expected to throw a ReleaseMilitaryException
        which should be caught from the caller of allocate() and trigger the
        release of all military resources so they can be reassigned to higher
        priority targets.

        :param ratio: ratio of available resources to minimum allocation
        """
        if ratio < self._military_reset_ratio and self._allocation_helper.try_reset:
            raise ReleaseMilitaryException

    @property
    def nearby_empire_count(self):
        """The number of enemy empires within at most 2 jumps."""
        return self._allocation_helper.enemy_supply.get(self.sys_id, 0)

    @property
    def threat_bias(self):
        """A constant threat biases added additively to the calculated threat."""
        return self._allocation_helper.threat_bias

    @property
    def safety_factor(self):
        """A multiplicative factor for threat calculations"""
        return self._allocation_helper.safety_factor

    @property
    def assigned_rating(self):
        """The combined rating of existing missions assigned to the system."""
        return self._allocation_helper.already_assigned_rating.get(self.sys_id, 0)

    @property
    def assigned_rating_vs_planets(self):
        return self._allocation_helper.already_assigned_rating_vs_planets.get(self.sys_id, 0)

    def _local_threat(self):
        """Military rating of enemies present in the system."""
        return get_system_local_threat(self.sys_id)

    def _neighbor_threat(self):
        """Military rating of enemies present in neighboring system."""
        return get_system_neighbor_threat(self.sys_id)

    def _jump2_threat(self):
        """Military rating of enemies present 2 jumps away from the system."""
        return get_system_jump2_threat(self.sys_id)

    def _potential_threat(self):
        """Number of nearby enemies times the average enemy rating weighted by _potential_threat_factor"""
        return self.nearby_empire_count * enemy_rating() * self._potential_threat_factor

    def _regional_threat(self):
        """Threat derived from enemy supply lanes."""
        return get_system_regional_threat(self.sys_id)

    def _potential_support(self):
        """Military rating of our forces in neighboring systems."""
        return get_system_neighbor_support(self.sys_id)

    def _planet_threat(self):
        return get_system_planetary_threat(self.sys_id)

    def _enemy_ship_count(self):
        return get_aistate().systemStatus.get(self.sys_id, {}).get("enemy_ship_count", 0.0)


class CapitalDefenseAllocator(Allocator):
    _allocation_group = "capitol"
    _military_reset_ratio = 0.5

    def _minimum_allocation(self, threat):
        nearby_forces = combine_ratings(self.assigned_rating, self._potential_support())
        return max(
            rating_needed(self._regional_threat(), nearby_forces), rating_needed(1.4 * threat, self.assigned_rating)
        )

    def _maximum_allocation(self, threat):
        return max(
            rating_needed(1.5 * self._regional_threat(), self.assigned_rating),
            rating_needed(2 * threat, self.assigned_rating),
        )

    def _calculate_threat(self):
        potential_threat = max(self._potential_threat() - self._potential_support(), 0)
        actual_threat = self.safety_factor * (
            2 * self.threat_bias + +combine_ratings(self._local_threat(), self._neighbor_threat())
        )
        return potential_threat + actual_threat

    def _take_any(self):
        return True


class PlanetDefenseAllocator(Allocator):
    _allocation_group = "occupied"
    _min_alloc_factor = 1.1
    _max_alloc_factor = 1.5
    _potential_threat_factor = 0.5
    _military_reset_ratio = 0.8

    def allocate(self):
        remaining_rating = self._allocation_helper.remaining_rating
        if remaining_rating > 0:
            super().allocate()
            return
        if self._minimum_allocation(self._calculate_threat()):
            pass  # raise ReleaseMilitaryException TODO

    def _minimum_allocation(self, threat):
        super_call = super()._minimum_allocation(threat)
        restriction = 0.5 * self._allocation_helper.available_rating
        return min(super_call, restriction)

    def _calculate_threat(self):
        nearby_forces = combine_ratings(self._potential_support(), self.assigned_rating)
        return (
            self.threat_bias
            + +self.safety_factor * combine_ratings(self._local_threat(), self._neighbor_threat())
            + +max(0.0, self._potential_threat() + self._jump2_threat() - nearby_forces)
        )

    def _take_any(self):
        return True


class TargetAllocator(Allocator):
    _allocation_group = "otherTargets"
    _min_alloc_factor = 1.3
    _max_alloc_factor = 2.5
    _potential_threat_factor = 0.5

    def _calculate_threat(self):
        return (
            self.threat_bias
            + +self.safety_factor
            * combine_ratings(self._local_threat(), 0.75 * self._neighbor_threat(), 0.5 * self._jump2_threat())
            + self._potential_threat()
        )

    def _take_any(self):
        return self.assigned_rating > 0

    def _planet_threat_multiplier(self):
        # to the extent that enemy planetary defenses are bolstered by fleet defenses, a smaller portion of our
        # attacks will land on the planet and hence we need a greater portion of planet-effective attacks.  One
        # desired characteristic of the following planet_threat_multiplier is that if the entire local threat is
        # due to planetary threat then we want the multiplier to be unity.  Furthermore, the more enemy ships are
        # present, the smaller proportion of our attacks would be directed against the enemy planet.  The following is
        # just one of many forms of calculation that might work reasonably.
        # TODO: assess and revamp the planet_threat_multiplier calculation
        return (
            (self._enemy_ship_count() + self._local_threat() / self._planet_threat()) ** 0.5
            if self._planet_threat() > 0
            else 1.0
        )

    def _allocation_vs_planets(self):
        return rating_needed(
            self.safety_factor * self._planet_threat_multiplier() * self._planet_threat(),
            self.assigned_rating_vs_planets,
        )


class TopTargetAllocator(TargetAllocator):
    _allocation_group = "topTargets"
    _max_alloc_factor = 3


class OutpostTargetAllocator(TargetAllocator):
    _max_alloc_factor = 3


class LocalThreatAllocator(Allocator):
    _potential_threat_factor = 0
    _min_alloc_factor = 1.3
    _max_alloc_factor = 2
    _allocation_group = "otherTargets"

    def _calculate_threat(self):
        systems_status = get_aistate().systemStatus.get(self.sys_id, {})
        threat = self.safety_factor * combine_ratings(
            systems_status.get("fleetThreat", 0),
            systems_status.get("monsterThreat", 0) + +systems_status.get("planetThreat", 0),
        )
        return self.threat_bias + threat

    def _take_any(self):
        return False


class InteriorTargetsAllocator(LocalThreatAllocator):
    _max_alloc_factor = 2.5
    _min_alloc_factor = 1.3

    def _calculate_threat(self):
        return self.threat_bias + self.safety_factor * self._local_threat()

    def _maximum_allocation(self, threat):
        return self._max_alloc_factor * min(self._minimum_allocation(threat), self._allocation_helper.remaining_rating)

    def _take_any(self):
        return self.assigned_rating > 0


class BorderSecurityAllocator(LocalThreatAllocator):
    _min_alloc_factor = 1.2
    _max_alloc_factor = 2
    _allocation_group = "accessibleTargets"

    def __init__(self, sys_id, allocation_helper):
        super().__init__(sys_id, allocation_helper)

    def _maximum_allocation(self, threat):
        return self._max_alloc_factor * self.safety_factor * max(self._local_threat(), self._neighbor_threat())


class ReleaseMilitaryException(Exception):
    pass


# TODO: May want to move these functions into AIstate class
def get_system_local_threat(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("totalThreat", 0.0)


def get_system_jump2_threat(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("jump2_threat", 0.0)


def get_system_neighbor_support(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("my_neighbor_rating", 0.0)


def get_system_neighbor_threat(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("neighborThreat", 0.0)


def get_system_regional_threat(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("regional_threat", 0.0)


def get_system_planetary_threat(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("planetThreat", 0.0)


def enemy_rating() -> float:
    return get_aistate().empire_standard_enemy_rating


def get_my_defense_rating_in_system(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("mydefenses", {}).get("overall")


def enemies_nearly_supplying_system(sys_id):
    return get_aistate().systemStatus.get(sys_id, {}).get("enemies_nearly_supplied", [])


def get_military_fleets(mil_fleets_ids=None, try_reset=True, thisround="Main"):  # noqa: C901
    """Get armed military fleets."""
    global _military_allocations

    universe = fo.getUniverse()
    home_system_id = PlanetUtilsAI.get_capital_sys_id()

    all_military_fleet_ids = (
        mil_fleets_ids
        if mil_fleets_ids is not None
        else FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
    )

    mil_fleets_ids = list(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_military_fleet_ids))
    mil_needing_repair_ids, mil_fleets_ids = avail_mil_needing_repair(mil_fleets_ids, split_ships=True)
    avail_mil_rating = combine_ratings(get_fleet_rating(x) for x in mil_fleets_ids)

    if not mil_fleets_ids:
        if "Main" in thisround:
            _military_allocations = []
        return []

    # for each system, get total rating of fleets assigned to it
    already_assigned_rating = {}
    already_assigned_rating_vs_planets = {}
    aistate = get_aistate()
    systems_status = aistate.systemStatus
    enemy_sup_factor = {}  # enemy supply
    for sys_id in universe.systemIDs:
        already_assigned_rating[sys_id] = 0
        already_assigned_rating_vs_planets[sys_id] = 0
        enemy_sup_factor[sys_id] = min(2, len(systems_status.get(sys_id, {}).get("enemies_nearly_supplied", [])))
    for fleet_id in [fid for fid in all_military_fleet_ids if fid not in mil_fleets_ids]:
        ai_fleet_mission = aistate.get_fleet_mission(fleet_id)
        if not ai_fleet_mission.target:  # shouldn't really be possible
            continue
        last_sys = (
            ai_fleet_mission.target.get_system().id
        )  # will count this fleet as assigned to last system in target list  # TODO last_sys or target sys?
        this_rating = get_fleet_rating(fleet_id)
        this_rating_vs_planets = get_fleet_rating_against_planets(fleet_id)
        already_assigned_rating[last_sys] = combine_ratings(already_assigned_rating.get(last_sys, 0), this_rating)
        already_assigned_rating_vs_planets[last_sys] = combine_ratings(
            already_assigned_rating_vs_planets.get(last_sys, 0), this_rating_vs_planets
        )
    for sys_id in universe.systemIDs:
        my_defense_rating = systems_status.get(sys_id, {}).get("mydefenses", {}).get("overall", 0)
        already_assigned_rating[sys_id] = combine_ratings(my_defense_rating, already_assigned_rating[sys_id])
        if _verbose_mil_reporting and already_assigned_rating[sys_id]:
            debug(
                f"\t System {universe.getSystem(sys_id)} already assigned rating {already_assigned_rating[sys_id]:.1f}"
            )

    # get systems to defend
    capital_id = PlanetUtilsAI.get_capital()
    if capital_id is not None:
        capital_planet = universe.getPlanet(capital_id)
    else:
        capital_planet = None
    # TODO: if no owned planets try to capture one!
    if capital_planet:
        capital_sys_id = capital_planet.systemID
    else:  # should be rare, but so as to not break code below, pick a randomish mil-centroid system
        capital_sys_id = None  # unless we can find one to use
        system_dict = {}
        for fleet_id in all_military_fleet_ids:
            status = aistate.fleetStatus.get(fleet_id, None)
            if status is not None:
                system_id = status["sysID"]
                if not list(universe.getSystem(system_id).planetIDs):
                    continue
                system_dict[system_id] = system_dict.get(system_id, 0) + status.get("rating", 0)
        ranked_systems = sorted([(val, sys_id) for sys_id, val in system_dict.items()])
        if ranked_systems:
            capital_sys_id = ranked_systems[-1][-1]
        else:
            try:
                capital_sys_id = next(iter(aistate.fleetStatus.items()))[1]["sysID"]
            except:  # noqa: E722
                pass

    num_targets = max(10, PriorityAI.allotted_outpost_targets)
    top_target_planets = (
        [
            pid
            for pid, pscore, trp in AIstate.invasionTargets[: PriorityAI.allotted_invasion_targets()]
            if pscore > InvasionAI.MIN_INVASION_SCORE
        ]
        + [
            pid
            for pid, (pscore, spec) in list(get_colonisable_outpost_ids(True).items())[:num_targets]
            if pscore > InvasionAI.MIN_INVASION_SCORE
        ]
        + [
            pid
            for pid, (pscore, spec) in list(get_colonisable_planet_ids(True).items())[:num_targets]
            if pscore > InvasionAI.MIN_INVASION_SCORE
        ]
    )
    top_target_planets.extend(aistate.qualifyingTroopBaseTargets.keys())

    base_col_target_systems = PlanetUtilsAI.get_systems(top_target_planets)
    top_target_systems = []
    for sys_id in AIstate.invasionTargetedSystemIDs + base_col_target_systems:
        if sys_id not in top_target_systems:
            if aistate.systemStatus[sys_id]["totalThreat"] > get_tot_mil_rating():
                continue
            top_target_systems.append(sys_id)  # doing this rather than set, to preserve order

    try:
        # capital defense
        allocation_helper = AllocationHelper(
            already_assigned_rating, already_assigned_rating_vs_planets, avail_mil_rating, try_reset
        )
        if capital_sys_id is not None:
            CapitalDefenseAllocator(capital_sys_id, allocation_helper).allocate()

        # defend other planets
        empire_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
        empire_occupied_system_ids = list(set(PlanetUtilsAI.get_systems(empire_planet_ids)) - {capital_sys_id})
        for sys_id in empire_occupied_system_ids:
            PlanetDefenseAllocator(sys_id, allocation_helper).allocate()

        # attack / protect high priority targets
        for sys_id in top_target_systems:
            TopTargetAllocator(sys_id, allocation_helper).allocate()

        # enemy planets
        other_targeted_system_ids = [
            sys_id
            for sys_id in set(PlanetUtilsAI.get_systems(AIstate.opponentPlanetIDs))
            if sys_id not in top_target_systems
        ]
        for sys_id in other_targeted_system_ids:
            TargetAllocator(sys_id, allocation_helper).allocate()

        # colony / outpost targets
        other_targeted_system_ids = [
            sys_id
            for sys_id in list(set(AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs))
            if sys_id not in top_target_systems
        ]
        for sys_id in other_targeted_system_ids:
            OutpostTargetAllocator(sys_id, allocation_helper).allocate()

        # TODO blockade enemy systems

        # interior systems
        targetable_ids = set(get_systems_by_supply_tier(0))
        current_mil_systems = [sid for sid, _, _, _, _ in allocation_helper.allocations]
        interior_targets1 = targetable_ids.difference(current_mil_systems)
        interior_targets = [
            sid
            for sid in interior_targets1
            if (
                allocation_helper.threat_bias + systems_status.get(sid, {}).get("totalThreat", 0)
                > 0.8 * allocation_helper.already_assigned_rating.get(sid, 0)
            )
        ]
        for sys_id in interior_targets:
            InteriorTargetsAllocator(sys_id, allocation_helper).allocate()

        # TODO Exploration targets

        # border protections
        visible_system_ids = aistate.visInteriorSystemIDs | aistate.visBorderSystemIDs
        accessible_system_ids = (
            [sys_id for sys_id in visible_system_ids if systems_connected(sys_id, home_system_id)]
            if home_system_id != INVALID_ID
            else []
        )
        current_mil_systems = [sid for sid, alloc, rvp, take_any, _ in allocation_helper.allocations if alloc > 0]
        border_targets1 = [sid for sid in accessible_system_ids if sid not in current_mil_systems]
        border_targets = [
            sid
            for sid in border_targets1
            if (
                allocation_helper.threat_bias
                + systems_status.get(sid, {}).get("fleetThreat", 0)
                + systems_status.get(sid, {}).get("planetThreat", 0)
                > 0.8 * allocation_helper.already_assigned_rating[sid]
            )
        ]
        for sys_id in border_targets:
            BorderSecurityAllocator(sys_id, allocation_helper).allocate()
    except ReleaseMilitaryException:
        try_again(all_military_fleet_ids)
        return

    new_allocations = []
    remaining_mil_rating = avail_mil_rating
    # for top categories assign max_alloc right away as available
    for cat in ["capitol", "occupied", "topTargets"]:
        for sid, alloc, rvp, take_any, max_alloc in allocation_helper.allocation_by_groups.get(cat, []):
            if remaining_mil_rating <= 0:
                break
            this_alloc = min(remaining_mil_rating, max_alloc)
            new_allocations.append((sid, this_alloc, alloc, rvp, take_any))
            remaining_mil_rating = rating_difference(remaining_mil_rating, this_alloc)

    base_allocs = set()
    # for lower priority categories, first assign base_alloc around to all, then top up as available
    for cat in ["otherTargets", "accessibleTargets", "exploreTargets"]:
        for sid, alloc, rvp, take_any, max_alloc in allocation_helper.allocation_by_groups.get(cat, []):
            if remaining_mil_rating <= 0:
                break
            alloc = min(remaining_mil_rating, alloc)
            base_allocs.add(sid)
            remaining_mil_rating = rating_difference(remaining_mil_rating, alloc)
    for cat in ["otherTargets", "accessibleTargets", "exploreTargets"]:
        for sid, alloc, rvp, take_any, max_alloc in allocation_helper.allocation_by_groups.get(cat, []):
            if sid not in base_allocs:
                break
            if remaining_mil_rating <= 0:
                new_allocations.append((sid, alloc, alloc, rvp, take_any))
            else:
                local_max_avail = combine_ratings(remaining_mil_rating, alloc)
                new_rating = min(local_max_avail, max_alloc)
                new_allocations.append((sid, new_rating, alloc, rvp, take_any))
                remaining_mil_rating = rating_difference(local_max_avail, new_rating)

    if "Main" in thisround:
        _military_allocations = new_allocations
    if _verbose_mil_reporting or "Main" in thisround:
        debug(
            f"------------------------------\nFinal {thisround} Round Military Allocations: {{sid: alloc for sid, alloc, _, _, _ in new_allocations}} \n-----------------------"
        )
        debug("(Apparently) remaining military rating: %.1f" % remaining_mil_rating)

    return new_allocations


def assign_military_fleets_to_systems(use_fleet_id_list=None, allocations=None, round=1):  # noqa: C901
    # assign military fleets to military theater systems
    global _military_allocations
    universe = fo.getUniverse()
    if allocations is None:
        allocations = []

    doing_main = use_fleet_id_list is None
    aistate = get_aistate()
    if doing_main:
        aistate.misc["ReassignedFleetMissions"] = []
        base_defense_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.ORBITAL_DEFENSE)
        unassigned_base_defense_ids = FleetUtilsAI.extract_fleet_ids_without_mission_types(base_defense_ids)
        for fleet_id in unassigned_base_defense_ids:
            fleet = universe.getFleet(fleet_id)
            if not fleet:
                continue
            sys_id = fleet.systemID
            target = TargetSystem(sys_id)
            fleet_mission = aistate.get_fleet_mission(fleet_id)
            fleet_mission.clear_fleet_orders()
            fleet_mission.clear_target()
            mission_type = MissionType.ORBITAL_DEFENSE
            fleet_mission.set_target(mission_type, target)

        all_military_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
        if not all_military_fleet_ids:
            _military_allocations = []
            return
        avail_mil_fleet_ids = list(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_military_fleet_ids))
        mil_needing_repair_ids, avail_mil_fleet_ids = avail_mil_needing_repair(avail_mil_fleet_ids)
        these_allocations = _military_allocations
        debug("==================================================")
        debug("Assigning military fleets")
        debug("---------------------------------")
    else:
        avail_mil_fleet_ids = list(use_fleet_id_list)
        mil_needing_repair_ids, avail_mil_fleet_ids = avail_mil_needing_repair(avail_mil_fleet_ids)
        these_allocations = allocations

    # send_for_repair(mil_needing_repair_ids) #currently, let get taken care of by AIFleetMission.generate_fleet_orders()

    # get systems to defend

    avail_mil_fleet_ids = set(avail_mil_fleet_ids)
    for sys_id, alloc, minalloc, rvp, takeAny in these_allocations:
        if not doing_main and not avail_mil_fleet_ids:
            break
        debug("Allocating for: %s", TargetSystem(sys_id))
        found_fleets = []
        found_stats = {}
        ensure_return = sys_id not in set(
            AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs + AIstate.invasionTargetedSystemIDs
        )
        these_fleets = FleetUtilsAI.get_fleets_for_mission(
            target_stats={"rating": alloc, "ratingVsPlanets": rvp, "target_system": TargetSystem(sys_id)},
            min_stats={"rating": minalloc, "ratingVsPlanets": rvp, "target_system": TargetSystem(sys_id)},
            cur_stats=found_stats,
            starting_system=sys_id,
            fleet_pool_set=avail_mil_fleet_ids,
            fleet_list=found_fleets,
            ensure_return=ensure_return,
        )
        if not these_fleets:
            debug("Could not allocate any fleets.")
            if not found_fleets or not (FleetUtilsAI.stats_meet_reqs(found_stats, {"rating": minalloc}) or takeAny):
                if doing_main:
                    if _verbose_mil_reporting:
                        debug(
                            "NO available/suitable military allocation for system %d ( %s ) "
                            "-- requested allocation %8d, found available rating %8d in fleets %s"
                            % (
                                sys_id,
                                universe.getSystem(sys_id).name,
                                minalloc,
                                found_stats.get("rating", 0),
                                found_fleets,
                            )
                        )
                avail_mil_fleet_ids.update(found_fleets)
                continue
            else:
                these_fleets = found_fleets
        else:
            debug("Assigning fleets %s to target %s", these_fleets, TargetSystem(sys_id))
            if doing_main and _verbose_mil_reporting:
                debug(
                    "FULL+ military allocation for system %d ( %s )"
                    " -- requested allocation %8d, got %8d with fleets %s"
                    % (sys_id, universe.getSystem(sys_id).name, alloc, found_stats.get("rating", 0), these_fleets)
                )
        target = TargetSystem(sys_id)
        for fleet_id in these_fleets:
            fo.issueAggressionOrder(fleet_id, True)
            fleet_mission = aistate.get_fleet_mission(fleet_id)
            fleet_mission.clear_fleet_orders()
            fleet_mission.clear_target()
            if sys_id in set(
                AIstate.colonyTargetedSystemIDs + AIstate.outpostTargetedSystemIDs + AIstate.invasionTargetedSystemIDs
            ):
                mission_type = MissionType.SECURE
            elif get_owned_planets_in_system(sys_id):
                mission_type = MissionType.PROTECT_REGION
            else:
                mission_type = MissionType.MILITARY
            fleet_mission.set_target(mission_type, target)
            fleet_mission.generate_fleet_orders()
            if not doing_main:
                aistate.misc.setdefault("ReassignedFleetMissions", []).append(fleet_mission)

    if doing_main:
        debug("---------------------------------")
    last_round = 3
    last_round_name = "LastRound"
    if round <= last_round:
        # check if any fleets remain unassigned
        all_military_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
        avail_mil_fleet_ids = list(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_military_fleet_ids))
        allocations = []
        round += 1
        thisround = "Extras Remaining Round %d" % round if round < last_round else last_round_name
        if avail_mil_fleet_ids:
            debug("Round %s - still have available military fleets: %s", thisround, avail_mil_fleet_ids)
            allocations = get_military_fleets(mil_fleets_ids=avail_mil_fleet_ids, try_reset=False, thisround=thisround)
        if allocations:
            assign_military_fleets_to_systems(
                use_fleet_id_list=avail_mil_fleet_ids, allocations=allocations, round=round
            )
        else:
            # assign remaining fleets to nearest systems to protect.
            all_military_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
            avail_mil_fleet_ids = list(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_military_fleet_ids))

            def system_score(_fid, _sys_id):
                """Helper function to rank systems by priority"""
                jump_distance = universe.jumpDistance(_fid, _sys_id)
                if get_system_local_threat(_sys_id):
                    weight = 10
                elif get_system_neighbor_threat(_sys_id):
                    weight = 3
                elif get_system_jump2_threat(_sys_id):
                    weight = 1
                else:
                    weight = 1 / max(0.5, float(get_distance_to_enemy_supply(_sys_id))) ** 1.25
                return float(weight) / (jump_distance + 1)

            for fid in avail_mil_fleet_ids:
                fleet = universe.getFleet(fid)
                FleetUtilsAI.get_fleet_system(fleet)
                systems = get_owned_planets().keys()
                if not systems:
                    continue
                sys_id = max(systems, key=lambda x: system_score(fid, x))

                debug("Assigning leftover %s to system %d " "- nothing better to do.", fleet, sys_id)

                fleet_mission = aistate.get_fleet_mission(fid)
                fleet_mission.clear_fleet_orders()
                target_system = TargetSystem(sys_id)
                fleet_mission.set_target(MissionType.PROTECT_REGION, target_system)
                fleet_mission.generate_fleet_orders()


@cache_by_turn_persistent
def get_tot_mil_rating() -> float:
    """
    Give an assessment of total military rating considering all fleets as if distributed to separate systems.

    :return: a military rating value
    """
    return round(
        sum(get_fleet_rating(fleet_id) for fleet_id in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)),
        0,
    )


@cache_by_turn_persistent
def get_concentrated_tot_mil_rating() -> float:
    """
    Give an assessment of total military rating as if all fleets were merged into a single mega-fleet.

    :return: a military rating value
    """
    return round(
        combine_ratings(
            get_fleet_rating(fleet_id) for fleet_id in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
        ),
        0,
    )


@cache_by_turn_persistent
def get_num_military_ships():
    fleet_status = get_aistate().fleetStatus
    return sum(
        fleet_status.get(fid, {}).get("nships", 0)
        for fid in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
    )


def get_military_fleets_with_target_system(target_system_id):
    military_mission_types = [MissionType.MILITARY, MissionType.SECURE]
    found_fleets = []
    for fleet_mission in get_aistate().get_fleet_missions_with_any_mission_types(military_mission_types):
        if fleet_mission.target and fleet_mission.target.id == target_system_id:
            found_fleets.append(fleet_mission.fleet.id)
    return found_fleets



import freeOrionAIInterface as fo

from aistate_interface import get_aistate
from CombatRatingsAI import ShipCombatStats


class AdditionalSpecifications:
    """This class is a container for all kind of additional information
    and requirements we may want to use when assessing ship designs.

    methods for external use:
    convert_to_tuple(): Builds and returns a tuple of the class attributes
    update_enemy(enemy): updates enemy stats
    """

    def __init__(self):
        # TODO: Extend this framework according to needs of future implementations
        self.minimum_fuel = 0
        self.minimum_speed = 0
        self.orbital = False
        self.minimum_structure = 1
        self.minimum_fighter_launch_rate = 0
        self.enemy_shields = 1  # to avoid spamming flak cannons
        self.max_enemy_weapon_strength = 0
        self.avg_enemy_weapon_strength = 0
        self.expected_turns_till_fight = 2
        current_turn = fo.currentTurn()
        if current_turn < 80:
            self.enemy_mine_dmg = 0  # TODO: Implement the detection of actual enemy mine damage
        elif current_turn < 150:
            self.enemy_mine_dmg = 2
        elif current_turn < 230:
            self.enemy_mine_dmg = 6
        else:
            self.enemy_mine_dmg = 14
        self.enemy = None
        self.update_enemy(get_aistate().get_standard_enemy())

    def update_enemy(self, enemy: ShipCombatStats):
        """Read out the enemies stats and save them."""
        self.enemy = enemy
        self.enemy_shields = enemy.shields
        self.enemy_shields += 1  # add bias against weak weapons to account to allow weapons to stay longer relevant.

        enemy_attack_stats = enemy.attacks
        if enemy_attack_stats:
            self.max_enemy_weapon_strength = max(enemy_attack_stats.keys())
            n = 0
            d = 0
            for dmg, count in enemy_attack_stats.items():
                d += dmg * count
                n += count
            self.avg_enemy_weapon_strength = d // n  # TODO check if we need floor division here

    def convert_to_tuple(self):
        """Create a tuple of this class' attributes (e.g. to use as key in dict).

        :returns: tuple (minFuel,minSpeed,enemyDmg,enemyShield,enemyMineDmg)
        """
        return (
            "minFuel: %s" % self.minimum_fuel,
            "minSpeed: %s" % self.minimum_speed,
            "enemyDmg: %s" % self.max_enemy_weapon_strength,
            "enemyShields: %s" % self.enemy_shields,
            "enemyMineDmg: %s" % self.enemy_mine_dmg,
        )



from focs._effects import (
    EffectsGroup,
    IsSource,
    SetEmpireMeter,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_TRANS_ARCH",
    description="CON_TRANS_ARCH_DESC",
    short_description="VARIOUS_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=1000 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=[
        "CON_ORGANIC_STRC",
        "CON_NDIM_STRC",
    ],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_MEGALITH"),
        Item(type=UnlockPolicy, name="PLC_DIVINE_AUTHORITY"),
    ],
    effectsgroups=[
        EffectsGroup(
            scope=IsSource,
            effects=SetEmpireMeter(empire=Source.Owner, meter="SOCIAL_CATEGORY_NUM_POLICY_SLOTS", value=Value + 1),
        )
    ],
    graphic="icons/tech/transcendent_architecture.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BARREN_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import GREAT_INDUSTRY
from species.species_macros.influence import GOOD_INFLUENCE
from species.species_macros.native_fortification import ADVANCED_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_OFFENSE_TROOPS, GOOD_DEFENSE_TROOPS

Species(
    name="SP_KOBUNTURA",
    description="SP_KOBUNTURA_DESC",
    gameplay_description="SP_KOBUNTURA_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "SELF_SUSTAINING",
        "GREAT_INDUSTRY",
        "BAD_HAPPINESS",
        "GOOD_INFLUENCE",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_SELF_SUSTAINING_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "MIMETIC_ALLOY_SPECIAL",
        "SHIMMER_SILK_SPECIAL",
        "SUPERCONDUCTOR_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "GAIA_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_EXPLORATION",
        "PLC_ISOLATION",
        "PLC_INDUSTRIALISM",
        "PLC_INTERSTELLAR_INFRA",
    ],
    dislikes=[
        "FOCUS_RESEARCH",
        "FOCUS_INFLUENCE",
        "BLD_SHIPYARD_BASE",
        "BLD_SHIPYARD_ORBITAL_DRYDOCK",
        "BLD_SHIPYARD_CON_NANOROBO",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_CON_ADV_ENGINE",
        "KRAKEN_NEST_SPECIAL",
        "PANOPTICON_SPECIAL",
        "FRUIT_SPECIAL",
        "ELERIUM_SPECIAL",
        "PLC_CENTRALIZATION",
        "PLC_CHECKPOINTS",
        "PLC_ENVIRONMENTALISM",
        "PLC_NO_SUPPLY",
        "PLC_MODERATION",
    ],
    effectsgroups=[
        *GREAT_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GOOD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_KOBUNTURA"),
        *AVERAGE_SUPPLY,
        *GOOD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        # not for description,
        *ADVANCED_NATIVE_DEFENSE,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=BARREN_NARROW_EP,
    graphic="icons/species/intangible-04.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_weapons.ship_weapons import (
    WEAPON_BASE_EFFECTS,
    WEAPON_UPGRADE_CAPACITY_EFFECTS,
)

AD_2_upgrade = 2
AD_3_upgrade = 3
AD_2_plus_3_upgrade = AD_2_upgrade + AD_3_upgrade

# In absense of shields this weapon can get almost as good as Plasma. Against shields its pretty useless
Tech(
    name="SHP_WEAPON_ARC_DISRUPTOR_1",
    description="SHP_WEAPON_ARC_DISRUPTOR_1_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=12 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_ROOT_AGGRESSION"],
    unlock=Item(type=UnlockShipPart, name="SR_ARC_DISRUPTOR"),
    effectsgroups=WEAPON_BASE_EFFECTS("SR_ARC_DISRUPTOR"),
    graphic="icons/ship_parts/pulse-laser-1.png",
)

Tech(
    name="SHP_WEAPON_ARC_DISRUPTOR_2",
    description="SHP_WEAPON_ARC_DISRUPTOR_2_DESC",
    short_description="SHIP_WEAPON_IMPROVE_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=96 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_WEAPON_ARC_DISRUPTOR_1"],
    effectsgroups=WEAPON_UPGRADE_CAPACITY_EFFECTS("SHP_WEAPON_ARC_DISRUPTOR_2", "SR_ARC_DISRUPTOR", AD_2_upgrade),
    graphic="icons/ship_parts/pulse-laser-2.png",
)


Tech(
    name="SHP_WEAPON_ARC_DISRUPTOR_3",
    description="SHP_WEAPON_ARC_DISRUPTOR_3_DESC",
    short_description="SHIP_WEAPON_IMPROVE_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=12,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_WEAPON_ARC_DISRUPTOR_2"],
    effectsgroups=WEAPON_UPGRADE_CAPACITY_EFFECTS(
        "SHP_WEAPON_ARC_DISRUPTOR_3", "SR_ARC_DISRUPTOR", AD_3_upgrade, upgraded_damage_override=AD_2_plus_3_upgrade
    ),
    graphic="icons/ship_parts/pulse-laser-3.png",
)



import cProfile
import os
import pstats
import time
from io import StringIO


def profile(save_path, sort_by="cumulative"):
    """
    Profile function decorator.

    Each function execution will add new entry to file.

    Result stored in user directory in profile catalog. See lne in logs for exact position.

    :param save_path: path to save profile, for example:
        os.path.join(fo.getUserDataDir(), 'profiling', base_path, get_aistate().uid)
    :type save_path: str
    :param sort_by: sort stats https://docs.python.org/2/library/profile.html#pstats.Stats.sort_stats
    :type sort_by: str
    """

    def argument_wrapper(function):
        def wrapper(*args, **kwargs):
            pr = cProfile.Profile()
            pr.enable()
            start = time.clock()
            result = function(*args, **kwargs)
            end = time.clock()
            pr.disable()
            print(f"Profile {function.__name__} took {end - start:f} s, saved to {save_path}")
            s = StringIO()
            ps = pstats.Stats(pr, stream=s).strip_dirs().sort_stats(sort_by)
            ps.print_stats()

            base_path = os.path.dirname(save_path)
            if not os.path.exists(base_path):
                os.makedirs(base_path)
            with open(save_path, "a") as f:
                f.write(s.getvalue())
                f.write("\n")

            return result

        return wrapper

    return argument_wrapper



from focs._effects import (
    Conditional,
    EffectsGroup,
    LocalCandidate,
    MaxOf,
    MinOf,
    NamedReal,
    NamedRealLookup,
    OwnedBy,
    Planet,
    SetInfluence,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import AFTER_ALL_TARGET_MAX_METERS_PRIORITY

Tech(
    name="GRO_TRANSORG_SENT",
    description="GRO_TRANSORG_SENT_DESC",
    short_description="POLICY_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=126 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_GROWTH_CATEGORY", "THEORY"],
    prerequisites=["GRO_TERRAFORM"],
    unlock=Item(type=UnlockPolicy, name="PLC_TERRAFORMING"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=Conditional(
                condition=(Value(LocalCandidate.Influence) <= Value(LocalCandidate.TargetInfluence)),
                effects=[
                    SetInfluence(
                        value=MinOf(
                            float,
                            Value + NamedReal(name="GRO_TRANSORG_INFLUENCE_RATE", value=1.0),
                            Value(Target.TargetInfluence),
                        )
                    )
                ],
                else_=[
                    SetInfluence(
                        value=MaxOf(
                            float,
                            Value - NamedRealLookup(name="GRO_TRANSORG_INFLUENCE_RATE"),
                            Value(Target.TargetInfluence),
                        )
                    )
                ],
            ),
        )
    ],
    graphic="icons/tech/trans_organic_sentience.png",
)



from collections.abc import Iterable, Iterator
from typing import Optional


class UnmatchedType(Exception):
    """
    Excaption raised when type could not be matched.
    """


class Token:
    def __init__(self, value: str, head: str, tail: str, arg_count: int):
        self.value = value
        self.head = head
        self.tail = tail
        self.arg_count = arg_count

    def wrap(self, *args: str):
        return f"{self.head}{', '.join(args)}{self.tail}"

    def __len__(self):
        return len(self.value)


class ScalarToken(Token):
    def __init__(self, value: str, type_: str):
        super().__init__(value, type_, "", 0)


class SimpleCollectionToken(Token):
    def __init__(self, value: str, type_):
        super().__init__(value, f"{type_}[", "]", 1)


class PairToken(SimpleCollectionToken):
    def __init__(self, value: str):
        super().__init__(value, "tuple")

    def wrap(self, res: str):
        return f"{self.head}{res}, {res}{self.tail}"


class MapToken(Token):
    def __init__(self, value: str):
        super().__init__(value, "Mapping[", "]", 2)


_tokens: list[Token] = [
    SimpleCollectionToken("Vec", "Sequence"),
    SimpleCollectionToken("Set", "Set"),
    MapToken("Map"),
    PairToken("Pair"),
    ScalarToken("Int", "int"),
    ScalarToken("Bool", "bool"),
    ScalarToken("Visibility", "visibility"),
    ScalarToken("UnlockableItem", "UnlockableItem"),
    ScalarToken("MeterType", "meterType"),
    ScalarToken("AccountingInfo", "AccountingInfo"),
    ScalarToken("ShipPart", "shipPart"),
    ScalarToken("ShipSlotType", "shipSlotType"),
    ScalarToken("Meter", "meter"),
    ScalarToken("Float", "float"),
    ScalarToken("Double", "float"),
    ScalarToken("String", "str"),
]


def _split_once(string):
    for token in _tokens:
        if string.endswith(token.value):
            size = len(token)
            return string[:-size], token
    else:
        raise UnmatchedType(f"Unsupported string: {string}")


def _iter_arguments(string) -> Iterator[Token]:
    string = string.replace("_", "")
    while string:
        string, token = _split_once(string)
        yield token


def _fetch_group(stream: Iterable[Token]):
    first = next(stream)
    if first.arg_count == 0:
        return [first, []]
    if first.arg_count == 1:
        return [first, [_fetch_group(stream)]]
    if first.arg_count == 2:
        val = _fetch_group(stream)
        key = _fetch_group(stream)
        return [first, [key, val]]


def make_type(string: Optional[str]):
    if not string:
        return ""
    try:
        token, args = _fetch_group(_iter_arguments(string))
    except UnmatchedType:
        return string

    def wrap(token, args):
        if not args:
            return token.wrap("")
        else:
            new_args = [wrap(t, a) for t, a in args]
            return token.wrap(*new_args)

    return wrap(token, args)


def is_collection_type(type_name: str) -> bool:
    type_ = make_type(type_name)
    return type_.startswith(("Sequence[", "Set[", "Mapping[", "tuple["))



"""Handle diplomatic messages and response determination."""
import freeOrionAIInterface as fo
import random
from logging import debug

from aistate_interface import get_aistate
from character.character_module import Aggression
from character.character_strings_module import possible_greetings
from common.fo_typing import EmpireId
from common.option_tools import get_option_dict
from freeorion_tools.translation import UserStringList

gang_up_turn = int(get_option_dict().get("gang_up_turn", "99999"))


def check_gang_up():
    if fo.currentTurn() == gang_up_turn:
        for empire_id in fo.allEmpireIDs():
            player_id = fo.empirePlayerID(empire_id)
            if not fo.playerIsAI(player_id):
                fo.sendChatMessage(player_id, "Let's gang up to destroy this arrogant player!")
            elif empire_id > fo.empireID():
                offer = fo.diplomaticMessage(empire_id, fo.empireID(), fo.diplomaticMessageType.peaceProposal)
                debug(f"Sending diplomatic message to empire {empire_id} of type {offer.type}")
                fo.sendDiplomaticMessage(offer)


def get_diplomatic_status(empire_id: EmpireId) -> fo.diplomaticStatus:
    return fo.getDiplomaticStatus(fo.empireID(), empire_id)


def handle_pregame_chat(sender_player_id, message_txt):
    if fo.playerIsAI(sender_player_id):
        return
    possible_acknowledgments = UserStringList("AI_PREGAME_ACKNOWLEDGEMENTS__LIST")
    acknowledgement = random.choice(possible_acknowledgments)
    debug(
        "Acknowledging pregame chat with initial message (from %d choices): '%s'"
        % (len(possible_acknowledgments), acknowledgement)
    )
    fo.sendChatMessage(sender_player_id, acknowledgement)


class DiplomaticCorp:
    def __init__(self):
        self.be_chatty = True

    def handle_diplomatic_message(self, message):  # noqa: C901
        """Handle a diplomatic message update from the server,
        such as if another player declares war, accepts peace, or cancels a proposed peace treaty.
        :param message: message.recipient and message.sender are respective empire IDs
        :return:
        """
        debug(
            "Received diplomatic {} message from {} to {}.".format(
                message.type,
                fo.getEmpire(message.sender),
                "me" if message.recipient == fo.empireID() else fo.getEmpire(message.recipient),
            )
        )
        # TODO: remove the following early return once proper support for third party diplomatic history is added
        if message.recipient != fo.empireID():
            return
        aistate = get_aistate()
        if message.type == fo.diplomaticMessageType.alliesProposal:
            aistate.log_alliance_request(message.sender, message.recipient)
            proposal_sender_player = fo.empirePlayerID(message.sender)
            attitude = aistate.character.attitude_to_empire(message.sender, aistate.diplomatic_logs)
            possible_acknowledgments = []
            aggression = aistate.character.get_trait(Aggression)
            if fo.currentTurn() >= gang_up_turn and fo.playerIsAI(fo.empirePlayerID(message.sender)):
                accept_proposal = True
            elif aggression.key == fo.aggression.beginner:
                accept_proposal = True
            elif aggression.key == fo.aggression.turtle:
                accept_proposal = attitude > 0
            elif aggression.key == fo.aggression.cautious:
                accept_proposal = attitude > 5
            else:  # aggression typical or greater
                accept_proposal = False
            if aggression.key <= fo.aggression.typical:
                possible_acknowledgments = UserStringList("AI_ALLIANCE_PROPOSAL_ACKNOWLEDGEMENTS_MILD_LIST")
                if accept_proposal:
                    possible_replies = UserStringList("AI_ALLIANCE_PROPOSAL_RESPONSES_YES_MILD_LIST")
                else:
                    possible_replies = UserStringList("AI_ALLIANCE_PROPOSAL_RESPONSES_NO_MILD_LIST")
            else:
                possible_acknowledgments = UserStringList("AI_ALLIANCE_PROPOSAL_ACKNOWLEDGEMENTS_HARSH_LIST")
                if accept_proposal:
                    possible_replies = UserStringList("AI_ALLIANCE_PROPOSAL_RESPONSES_YES_HARSH_LIST")
                else:
                    possible_replies = UserStringList("AI_ALLIANCE_PROPOSAL_RESPONSES_NO_HARSH_LIST")
            acknowledgement = random.choice(possible_acknowledgments)
            reply_text = random.choice(possible_replies)
            debug(
                "Acknowledging proposal with initial message (from %d choices): '%s'"
                % (len(possible_acknowledgments), acknowledgement)
            )
            fo.sendChatMessage(proposal_sender_player, acknowledgement)
            if accept_proposal:
                diplo_reply = fo.diplomaticMessage(
                    message.recipient, message.sender, fo.diplomaticMessageType.acceptAlliesProposal
                )
                debug(f"Sending diplomatic message to empire {message.sender} of type {diplo_reply.type}")
                fo.sendDiplomaticMessage(diplo_reply)
            debug(
                "sending chat to player %d of empire %d, message body: '%s'"
                % (proposal_sender_player, message.sender, reply_text)
            )
            fo.sendChatMessage(proposal_sender_player, reply_text)
        if message.type == fo.diplomaticMessageType.peaceProposal:
            aistate.log_peace_request(message.sender, message.recipient)
            proposal_sender_player = fo.empirePlayerID(message.sender)
            attitude = aistate.character.attitude_to_empire(message.sender, aistate.diplomatic_logs)
            possible_acknowledgments = []
            aggression = aistate.character.get_trait(Aggression)
            if aggression.key <= fo.aggression.typical:
                possible_acknowledgments = UserStringList("AI_PEACE_PROPOSAL_ACKNOWLEDGEMENTS_MILD_LIST")
                if attitude > 0:
                    possible_replies = UserStringList("AI_PEACE_PROPOSAL_RESPONSES_YES_MILD_LIST")
                else:
                    possible_replies = UserStringList("AI_PEACE_PROPOSAL_RESPONSES_NO_MILD_LIST")
            else:
                possible_acknowledgments = UserStringList("AI_PEACE_PROPOSAL_ACKNOWLEDGEMENTS_HARSH_LIST")
                if attitude > 0:
                    possible_replies = UserStringList("AI_PEACE_PROPOSAL_RESPONSES_YES_HARSH_LIST")
                else:
                    possible_replies = UserStringList("AI_PEACE_PROPOSAL_RESPONSES_NO_HARSH_LIST")
            acknowledgement = random.choice(possible_acknowledgments)
            reply_text = random.choice(possible_replies)
            debug(
                "Acknowledging proposal with initial message (from %d choices): '%s'"
                % (len(possible_acknowledgments), acknowledgement)
            )
            fo.sendChatMessage(proposal_sender_player, acknowledgement)
            if fo.currentTurn() >= gang_up_turn and fo.playerIsAI(fo.empirePlayerID(message.sender)):
                reply = fo.diplomaticMessage(
                    message.recipient, message.sender, fo.diplomaticMessageType.acceptPeaceProposal
                )
                fo.sendDiplomaticMessage(reply)
                allies = fo.diplomaticMessage(
                    message.recipient, message.sender, fo.diplomaticMessageType.alliesProposal
                )
                fo.sendDiplomaticMessage(allies)
            elif attitude > 0:
                diplo_reply = fo.diplomaticMessage(
                    message.recipient, message.sender, fo.diplomaticMessageType.acceptPeaceProposal
                )
                debug(f"Sending diplomatic message to empire {message.sender} of type {diplo_reply.type}")
                fo.sendDiplomaticMessage(diplo_reply)
            debug(
                "sending chat to player %d of empire %d, message body: '%s'"
                % (proposal_sender_player, message.sender, reply_text)
            )
            fo.sendChatMessage(proposal_sender_player, reply_text)
        elif message.type == fo.diplomaticMessageType.warDeclaration:
            # note: apparently this is currently (normally?) sent not as a warDeclaration,
            # but as a simple diplomatic_status_update to war
            aistate.log_war_declaration(message.sender, message.recipient)

    @staticmethod
    def get_first_turn_greet_message():
        greets = possible_greetings(get_aistate().character)
        # no such entry
        if len(greets) == 1 and greets[0] == "?":
            greets = UserStringList("AI_FIRST_TURN_GREETING_BEGINNER")
        return random.choice(greets)

    def handle_diplomatic_status_update(self, status_update: fo.diplomaticStatusUpdate):
        """Handle an update about the diplomatic status between players, which may
        or may not include this player."""
        debug(
            f"Received diplomatic status update to {status_update.status} about empire {status_update.empire1} and empire {status_update.empire2}"
        )
        if status_update.empire2 == fo.empireID() and status_update.status == fo.diplomaticStatus.war:
            get_aistate().log_war_declaration(status_update.empire1, status_update.empire2)

    def handle_midgame_chat(self, sender_player_id, message_txt):
        debug("Midgame chat received from player %d, message: %s" % (sender_player_id, message_txt))
        if fo.playerIsAI(sender_player_id) or not self.be_chatty:
            return
        if "BE QUIET" in message_txt.upper():
            possible_acknowledgments = UserStringList("AI_BE_QUIET_ACKNOWLEDGEMENTS__LIST")
            acknowledgement = random.choice(possible_acknowledgments)
            debug(
                "Acknowledging 'Be Quiet' chat request with initial message (from %d choices): '%s'"
                % (len(possible_acknowledgments), acknowledgement)
            )
            fo.sendChatMessage(sender_player_id, acknowledgement)
            self.be_chatty = False
            return
        if random.random() > 0.25:
            return
        possible_acknowledgments = UserStringList("AI_MIDGAME_ACKNOWLEDGEMENTS__LIST")
        acknowledgement = random.choice(possible_acknowledgments)
        debug(
            "Acknowledging midgame chat with initial message (from %d choices): '%s'"
            % (len(possible_acknowledgments), acknowledgement)
        )
        fo.sendChatMessage(sender_player_id, acknowledgement)
        self.be_chatty = False


class BeginnerDiplomaticCorp(DiplomaticCorp):
    pass


class ManiacalDiplomaticCorp(DiplomaticCorp):
    def __init__(self):
        DiplomaticCorp.__init__(self)
        self.be_chatty = False



from focs._effects import (
    EffectsGroup,
    GalaxySize,
    GameRule,
    IsSource,
    Victory,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_TRANSCEND",
    description="LRN_TRANSCEND_DESC",
    short_description="VICTORY_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=100.0 * GalaxySize * GameRule(type=float, name="RULE_SINGULARITY_COST_FACTOR") * TECH_COST_MULTIPLIER,
    researchturns=20,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=[
        "CON_TRANS_ARCH",
        "LRN_PSY_DOM",
        "PRO_ZERO_GEN",
        "GRO_ENERGY_META",
    ],
    effectsgroups=[EffectsGroup(scope=IsSource, effects=[Victory(reason="VICTORY_TECH")])],
    graphic="icons/tech/singularity_of_transcendence.png",
)
